#ifndef PROCESS_CONST_H_
#define PROCESS_CONST_H_

#include <complex>
#include <string> // filename reading
#include "interfaces.h" // tred integral interfaces

class ProcessConst {
// Switch the typedefs below to change which libraries are used for integral
// interfaces. Options can be found in tred/interfaces.h.
  typedef OLInterface ScalarIntInterface_t;
  typedef TRDerivativeInterface DerivativeIntInterface_t;
  typedef std::complex<double> cmplx;
  typedef cmplx mass_t;

  public:
  static const double ALT_GAMMA_FAC;
  static const double N_COLORS;
  static const double ZERO_TEST;

  ProcessConst();
  ~ProcessConst();

  void compute(bool QCD, bool EWK);
  void computeEWRenorm();
  void computeEWRenorm2(double);
  void computeQCDRenorm();
  void computeQCDRenorm2(double);
  mass_t get_mass(const std::string & name);
  double get_param(const std::string & name);
  void load_params(const std::string & filename);
  // Returns the internal process/tred mass name corresponding to a parameter name/width
  static std::string mass_name(const std::string & name);
  int set_param(const std::string & name, double val);
  static void set_param_in_file(const std::string & name, double val,
                                const std::string & filename);
  void print_EW_CTs();
  void printUserParameters();

  // Returns complex mass squared given real mass and width
  cmplx complex_mass_sq(double mass, double width);
  // Function to turn a set of mass parameters (for one particle) into complex
  // for the complex-mass scheme--m and w are inputs
  void complexify_mass(double m_in, double w, cmplx & m,
                       cmplx & mst, cmplx & m2);
  // Function to turn masses into complex masses for the complex-mass scheme
  void complexify_masses();
  // Returns complex mass to sqrt(Re(m^2))
  double real_mass(cmplx mass);
  // Functions to reverse transition to complex pole mass
  void uncomplexify_mass(double m_in, cmplx & m, cmplx & mst, cmplx & m2);
  void uncomplexify_masses();
  // Safely updates mass, keeping renorm constants consistent with cm scheme
  void update_mass(double m_new, double & m_store, double w_store,
                   mass_t & m, mass_t & mst, mass_t & m2);
  void update_width(double w_new, double m_store, double & w_store,
                    mass_t & m, mass_t & mst, mass_t & m2);

  // Flag to determine whether masses are complexified before or after couplings
  // are determined from them: cmScheme = true means the masses are complexified
  // before the couplings are derived from them, such that a complex-mass scheme 
  // is used; cmScheme = false means the masses are complexified after the coup-
  // lings are derived, such that effectively only widths are used.
  bool cmScheme;

  // Flag to determine whether the Gmu scheme is used or not. This accounts to
  // saying dZe = dZe - 0.5*Dr, aka dZe|Gmu = dZe|alpha(0) - 0.5*Dr - If using
  // the Gmu scheme one has to pay attention to also use the correct input pa-
  // rameters and derive the correct derived parameters.
  bool GmuScheme;

  // QCD and EW couplings
  double e, alpha_e, g, alpha_s, GF;

  // Color correlation flags. Setting ccL0 to 1, while all others are set to 0., 
  // returns the Born, while setting ccLi,ccLj (i,j \neq 0) to 1. in pairs, 
  // while all others are set to 0., returns the <Ti*Tj> color correlated Born. 
  double ccL0,ccL1,ccL2,ccL3,ccL4,ccL5,ccL6,ccL7;

  // Inputs for masses, sqrt(Re(m^2))
  double mup_in, md_in, ms_in, mc_in, mb_in, mt_in;
  double me_in, mmu_in, mtau_in, mA_in, mH_in, mW_in, mZ_in;

  // Fermion masses
  mass_t mup, md, ms, mc, mb, mt, mup2, md2, ms2, mc2, mb2, mt2;
  mass_t me, mmu, mtau, me2, mmu2, mtau2;
  mass_t mupst, mdst, msst, mcst, mbst, mtst;
  mass_t mest, mmust, mtaust;
  mass_t mblogQ, mtlogQ;
  mass_t mblogG, mtlogG;

  // Boson masses
  mass_t mA, mH, mW, mZ, mA2, mH2, mW2, mZ2;
  mass_t mAst, mHst, mWst, mZst;

  // Widths
  double wup, wd, ws, wc, wb, wt;
  double we, wmu, wtau;
  double wA, wH, wW, wZ;

  // QED charges and weak isospins
  double Qu, Qd, Ql, T3u, T3d, T3l, T3n;

  // Weak mixing angle
  cmplx cw, cw2, sw, sw2;
  cmplx cwst, cw2st, swst, sw2st;

  // Z-fermion vector/axial-vector couplings
  cmplx vu, au, vd, ad, vl, al, vn, an, gfp, gfm;
  cmplx gup, gum, gdp, gdm, glp, glm, gnm;
  cmplx vust, aust, vdst, adst, vlst, alst, vnst, anst, gfpst, gfmst;
  cmplx gupst, gumst, gdpst, gdmst, glpst, glmst, gnmst;

  // Yukawa couplings
  cmplx yb, yt;
  cmplx ybst, ytst;

  // W,H mass as used in couplings, not in propagators
  cmplx mWcp, mWcpst, mHcp, mHcpst;

  // Renormalization scale and flavor counts
  double mu, musq;
  double nlf, nhf;
  double nlfprime, nhfprime;

  // Converts a factor of 1/Gamma(1-eps) to Gamma(1+eps)/eps if true
  bool use_alt_gamma_fac;

  // Other parameters
  double zerocutoff, soft_cut, sqrt2,ir_threshold;

  // Switch for HV scheme (1 = on, 0 = off)
  double HV1;

  // QCD ren. constants.
  // Finite pieces end with Q0.
  // Pole pieces end with Q1.
  cmplx dZxgsxQ0, dZxgsxQ1; // alpha_s
  cmplx dZxGxQ0, dZxGxQ1;   // gluons
  cmplx dZxqxQ0, dZxqxQ1;   // massless quarks
  cmplx dZxtxQ0, dZxtxQ1;   // top
  cmplx dZxmtxQ0, dZxmtxQ1; // top mass
  cmplx dZxbxQ0, dZxbxQ1;   // bottom
  cmplx dZxmbxQ0, dZxmbxQ1; // bottom mass

  // QCD LSZ factors (ext. w.f. ren.).
  cmplx dRxGxQ0, dRxGxQ1;   // gluons
  cmplx dRxqxQ0, dRxqxQ1;   // massless quarks
  cmplx dRxqmxQ0, dRxqmxQ1; // massive quarks
  cmplx dRxtxQ0, dRxtxQ1;   // massive bottom
  cmplx dRxbxQ0, dRxbxQ1;   // massive top

  // Electroweak ren. constants.
  // Finite pieces end with E0.
  // Pole pieces end with E1.
  cmplx dZfLlE0, dZfRlE0;
  cmplx dZfLnE0, dZfRnE0;
  cmplx dZfLiiE0, dZfRiiE0, dmfiE0;
  cmplx dZfLuuE0, dZfRuuE0;
  cmplx dZfLddE0, dZfRddE0;
  cmplx dZfLbbE0, dZfRbbE0, dmbE0;
  cmplx dZfLttE0, dZfRttE0, dmtE0;

  cmplx dZAAE0, dZAZE0, dZZAE0, dZZZE0, dZWE0, dZHE0;
  cmplx dmH2E0, dmW2E0, dmZ2E0, dZeE0, dswE0;

  cmplx sew0, sew0_pole;
  cmplx DrE0, DrE1;

  cmplx dZfLlE1, dZfRlE1;
  cmplx dZfLnE1, dZfRnE1;
  cmplx dZfLiiE1, dZfRiiE1, dmfiE1;
  cmplx dZfLuuE1, dZfRuuE1;
  cmplx dZfLddE1, dZfRddE1;
  cmplx dZfLbbE1, dZfRbbE1, dmbE1;
  cmplx dZfLttE1, dZfRttE1, dmtE1;

  cmplx dZfLlE0st, dZfRlE0st;
  cmplx dZfLnE0st, dZfRnE0st;
  cmplx dZfLiiE0st, dZfRiiE0st;
  cmplx dZfLuuE0st, dZfRuuE0st;
  cmplx dZfLddE0st, dZfRddE0st;
  cmplx dZfLbbE0st, dZfRbbE0st;
  cmplx dZfLttE0st, dZfRttE0st;

  cmplx dZfLlE1st, dZfRlE1st;
  cmplx dZfLnE1st, dZfRnE1st;
  cmplx dZfLiiE1st, dZfRiiE1st;
  cmplx dZfLuuE1st, dZfRuuE1st;
  cmplx dZfLddE1st, dZfRddE1st;
  cmplx dZfLbbE1st, dZfRbbE1st;
  cmplx dZfLttE1st, dZfRttE1st;

  cmplx dtE1, dtE0;
  cmplx dZAAE1, dZAZE1, dZZAE1, dZZZE1, dZWE1, dZHE1;
  cmplx dmH2E1, dmW2E1, dmZ2E1, dZeE1, dswE1;
  double EW_FAC;

  // Custom parameters
  // declarations: No custom parameters found.

  // Interfaces to libraries where we can find integral functions
  ScalarIntInterface_t scalar_if;
  DerivativeIntInterface_t derivative_if;

 private:  

  static const cmplx czero;
  static const cmplx cone;
};

inline ProcessConst::cmplx ProcessConst::complex_mass_sq(double mass,
                                                         double width) {
  return cmplx(mass*mass, -mass*width);
}

#endif // PROCESS_CONST_H_
