#ifndef COEFFICIENT_H_
#define COEFFICIENT_H_

#include <assert.h>
#include <vector>
#include <string>
#include <iostream>
#include <set>
#include <utility>
#include <map>
#include "common.h"
#include "kinematics.h"
#include "interfaces.h"

// Holds multiple pointers to invariants (now that we have multiple precision)

struct InvariantGroup {
    Invariants<fT1>::Ptr inv1;
#if MAX_PREC_LVL >= 2
    Invariants<fT2>::Ptr inv2;
#endif
#if MAX_PREC_LVL >= 3
    Invariants<fT3>::Ptr inv3;
#endif
};

/**********************************************************************
 *                                                                    *
 * Coefficient                                                        *
 *                                                                    *
 **********************************************************************/

// Holds state for each coefficient that is not dependent on the
// evaluation method, i.e. tensor indices, invariants etc. The final
// numerical value is *not* stored here, as different evaluation
// methods might in principle require to store different state on
// individual coefficients.
class Coefficient {
  public:
    typedef shared_ptr<Coefficient> Ptr;
    Coefficient(int points, InvariantGroup inv,
                const std::vector<int> & indices);

    int points() const;
    int rank() const;
    template <typename fT> typename Invariants<fT>::Ptr invariants() const;
    InvariantGroup invariant_group() const;
    const std::vector<int> & indices() const;

    int index_hash() const;
    bool is_scalar() const;
    bool has_gmunu() const;

    // Returns the number of indices with value i
    int n_index_value(int i) const;
    bool has_only_index(int i) const;

    Coefficient drop_index(int pos) const;
    Coefficient drop_indices(int start, int end) const;
    Coefficient prepend_index(int index);
    Coefficient append_index(int index);
    Coefficient replace_indices(int index);

    // Returns a coefficient with the n-th propagator dropped from
    // the original. NullCoefficient is returned if any of the
    // tensor indices {i_k} is equal to n. For n > 0 both the
    // invariants and the tensor indices in the returned
    // coefficient are corrected for the dropped propagator, while
    // for n == 0 only the invariants are modified (equivalent to
    // \tilde coefficient in DD)
    Coefficient::Ptr drop_propagator(int n) const;
    Coefficient::Ptr drop_propagator(std::multiset<int> n) const;

    void print(std::ostream & strm, std::string prefix = "") const;

    bool operator==(const Coefficient & coeff) const;

  private:
    int points_;
    std::vector<int> indices_;
    InvariantGroup inv_;
    int rank_;
};

/**********************************************************************
 *                                                                    *
 * Inlined functions                                                  *
 *                                                                    *
 **********************************************************************/

template <typename fT> inline typename Invariants<fT>::Ptr
Coefficient::invariants() const {
  std::cerr << "Unsupported evaluation level in Invariants." << std::endl;
  exit(1);
}

template <> inline typename Invariants<fT1>::Ptr
Coefficient::invariants<fT1>() const {
  return inv_.inv1;
}

#if MAX_PREC_LVL >= 2
template <> inline typename Invariants<fT2>::Ptr
Coefficient::invariants<fT2>() const {
  return inv_.inv2;
}
#endif
#if MAX_PREC_LVL >= 3
template <> inline typename Invariants<fT3>::Ptr
Coefficient::invariants<fT3>() const {
  return inv_.inv3;
}
#endif


inline bool Coefficient::operator==(const Coefficient & coeff) const {
  // TODO: should be able to return result of logical operation directly
  if (index_hash() == coeff.index_hash() &&
      invariants<fT1>()->get_hash() == coeff.invariants<fT1>()->get_hash())
    // check only first one, the rest should match
    return true;
  return false;
}

inline InvariantGroup Coefficient::invariant_group() const {
  return inv_;
}

inline const std::vector<int> & Coefficient::indices() const {
  return indices_;
}

inline int Coefficient::points() const {
  return points_;
}

inline int Coefficient::rank() const {
  return rank_;
}

inline bool Coefficient::has_gmunu() const {
  // TODO: should be able to return result of logical operation directly
  if (indices_.size() != 0 && indices_[0] == 0)
    return true;
  return false;
}

inline bool Coefficient::has_only_index(int i) const {
  for (std::vector<int>::const_iterator it = indices_.begin();
       it != indices_.end(); ++it)
    if (*it != i)
      return false;
  return true;
}

inline int Coefficient::n_index_value(int n) const {
  int count = 0;
  for (std::vector<int>::const_iterator i = indices_.begin();
       i != indices_.end(); ++i) {
    if (*i == n)
      count++;
    else if (*i > n)
      break;
  }
  return count;
}

inline bool Coefficient::is_scalar() const {
  return indices_.size() == 0;
}

#endif  // COEFFICIENT_H_
