#ifndef EA_NODE_H_
#define EA_NODE_H_

#include <map>
#include <vector>
#include "evaluation.h"
#include "tcomplex.h"

/**********************************************************************
 *                                                                    *
 * EANode                                                             *
 *                                                                    *
 **********************************************************************/

class EANode : public EvaluationNode {
  public:
    typedef shared_ptr<EANode> Ptr;
    EANode(const Coefficient & coeff, const CoefficientNode & cnode,
           EvaluationList* parent, CoefficientPool* pool);
    ~EANode();

    // Adds all dependencies of the current node to the node.
    // NOTE: This has side effects, as it adds to the
    // CoefficientPool
    void add_dependencies();
    void evaluate();

    typedef std::map<int, CoefficientNode::Ptr> IntNodeMap;
    typedef std::map<int, ContractedFirstProp*> IntCFPMap;
    typedef std::map<std::pair<int, int>, CoefficientNode::Ptr> PairIntNodeMap;

  private:
    void add_scalar_dependencies();
    void add_Eibar_dependencies(const std::vector<int> & inds,
                                int e_bar_index);
    void add_E00bar_dependencies(const std::vector<int> & inds,
                                 int e_bar_index);

    template <typename fT> void evaluateT(Poly3T<fT> & value);
    template <typename fT> Poly3T<fT> scalar_evaluate();
    template <typename fT> Poly3T<fT> e00_bar_evaluate(int index_id);
    template <typename fT> Poly3T<fT>
    ei_bar_evaluate(const std::vector<int> & inds, int index_id);

    bool coeff_has_permutations();
    // populates e_bar_indices_ with all possible swappings for
    // the evlaution of symmetric E coefficients from E_bar's
    void coeff_index_swappings();

  private:
    static const double term_cutoff_;
    // Indices of E_bar's needed to evaluate the value for coeff_.
    // First the index, second the multiplicity of equal
    // configurations. Imutable after construction.
    std::map<std::vector<int>, int> e_bar_indices_;

    // Dependencies
    //   for E_{i1...ip} with i1 != 0: d_dp0_ = D_{i2...ip}(0)
    //   for E_{00i2...ip}:            d_dp0_ = D_{00i2...ip}(0)
    IntCFPMap d_dp0_;

    // for E_{i1...ip} with i1 != 0: d_dpn_ = D_{i2...ip}(n)
    // for E_{00i2...ip}:            d_dpn_ = D_{00i2...ip}(n)
    std::map<int, IntNodeMap> d_dpn_;
    std::map<int, PairIntNodeMap> d_dpn_ai0_dik_;
    std::map<int, IntCFPMap> d_dp0_ai0_dik_;
};

#endif  // EA_NODE_H_
