#ifndef EVALUATION_H_
#define EVALUATION_H_

#include <iostream>
#include <string>
#include <algorithm>
#include <vector>
#include <list>
//#if __cplusplus >= 201103L
//#include <unordered_map>
//using std::unordered_map;
//#else
//#include <tr1/unordered_map>
//using std::tr1::unordered_map;
//#endif
#include "coefficient.h"
#include "cayleyzero_exception.h"
#include "methods.h" // definition for EvaluationsMethod enum

// switch this to change Scalar Interface Type;
// options are QLInterface (QCDLoop), QL2Interface (QCDLoop2),
// LTInterface (LoopTools), OLInterface (OneLOop)

/* The providers have certain advantages/disadvantages.
   --QCDLoop is fairly stable and standard, but does not support complex
   masses. Seems to have the best pole behavior, and overall the best
   stability across kinematic configurations.
   --QCDLoop2 is native C++, supports higher precision, but gives wrong answers
   for massive D's and possibly other things.
   --LoopTools is slow, and mostly useful for consistency checks if one is also
   using it to check tensor coefficients. Seems to have stable scalars.
   --OneLOop allows complex masses. It also seems to be most stable in
   the finite in many, but not all situations. In particular, the finite box
   seems very unstable for small masses (mb).
*/

typedef OLInterface ScalarIntInterface_t;

// switch this to change Tensor Interface Type (for debugging or pole check)
typedef PoleCheckInterface TensorIntInterface_t;
//typedef LTInterface TensorIntInterface_t;

/* The following three classes (and derivations from EvaluationNode
 * and EvaluationList) form an entity and are tightly coupled. The
 * user should only ever create an instance of CoefficientPool and
 * never use EvaluationList and EvaluationNode directly (except for
 * const access).
 *
 * Implementation note: One could declare the classes as mutual
 * friends and make their constructors private. This way, only const
 * references or const *::Ptr's are returned. Eventually, only the two
 * nontrivial CoefficientPool::create_coefficient() functions should
 * be public and return refs to the respective Coefficients. A
 * CoefficientPool compute() function than populates the Coefficients
 * in the set<> with the final results and they can be retrieved
 * through Coefficient::value() */

class CoefficientPool;
class EvaluationList;
class EvaluationNode;
class CoefficientNode;

class PVNode;
class ABNode;

// Handles child coefficients that result from contractions of the
// first propagator in the parent.
// Instead of calling Coefficient.drop_propagator(0) in the evaluation
// methods the parent coefficient is passed to a new instance of this
// class. The referenced CoefficientPool is then populated with all
// needed coefficients and the new EvaluationNodes for the
// pointer are stored. To evaluate the coefficient, the value() method
// is called, which is safe as children are always evaluated before
// their parents.
//
// Note: Construction of this class has side effects in
// CoefficientPool.
//
// Usage example:
//    In EvaluationNode::add_dependencies():
//        ContractedFirstProp cfp(coeff, *pool);
//    In EvaluationNode::evaluate():
//        Poly3 value = cfp.value();
class ContractedFirstProp {
  public:
    ContractedFirstProp(const Coefficient & coeff, CoefficientPool * pool,
                        bool isRescue);
    template <typename fT> Poly3T<fT> value() const;

    static void fill_indices(std::vector<std::vector<int> > & vec, int len,
                                 int min, int max);
  private:
    // Holds dependencies. Key values:
    //   key < 0:  special case A(0), -key is the power of the prefactor
    //   key == 0: invariant coefficient, map may not have more
    //             than one value and the value has to be of size one
    //   key > 0:  key-1 is summation index over children with same
    //             rank that are stored in the value.
    std::vector<std::pair<int,
                          std::vector<shared_ptr<CoefficientNode> > > > _deps;
    CoefficientPool* pool_;
};

/**********************************************************************
 *                                                                    *
 * EvaluationNode                                                     *
 *                                                                    *
 **********************************************************************/

class EvaluationNode {
  friend class ContractedFirstProp;
  public:
    typedef shared_ptr<EvaluationNode> Ptr;
    EvaluationNode(const Coefficient & coeff,
                   const CoefficientNode & cnode,
                   EvaluationList* parent,
                   CoefficientPool* pool);

    // Adds all dependencies of the current node to the node.
    // NOTE: This has side effects, as it adds to the CoefficientPool
    virtual void add_dependencies() = 0;

    // Note: evaluate() is virtual to accomodate different types of nodes.
    // However, now that we have different precision, the actual evaluation
    // is templated. Each type of node has identical code for evaluate() to
    // determine the precision to use in calling evaluateT<>() functions
    // unique to each type of node. Unfortunately this is necessary since
    // C++ doesn't allow virtual templated functions.
    virtual void evaluate() = 0;

    void check_pole(const Poly3T<fT1> & poly) const;
    const Coefficient & coeff() const;
    bool is_cayley_zero();
    Poly3T<double> value();
    template <typename fT> Poly3T<fT> valueT();
    virtual void clear();

    virtual void print() const;

    bool operator==(const EvaluationNode & node) const;
    bool operator==(const EvaluationNode & node);
    // check the canonical order for this type of node
    virtual bool cmp(const EvaluationNode & node) const;

  protected:
    typedef std::vector<EvaluationNode::Ptr> ptr_vector;

    void handle_instability(std::string nodename);

    const Coefficient & coeff_;
    // Reference to parent CoefficientNode that might call me
    const CoefficientNode & cnode_; // will probably need to make non-const
    EvaluationList* parent_;
    CoefficientPool* pool_;

    int evaluation_level_; // 0 for unevaluated
    static double instab_thresh_;
    static double fail_thresh_;
    bool is_cayley_zero_;
    Poly3T<fT1> value1_;
#if MAX_PREC_LVL >= 2
    Poly3T<fT2> value2_;
#if MAX_PREC_LVL >= 3
    Poly3T<fT3> value3_;
#endif
#endif
    void not_eval_msg() const;
};

// CoefficientNode is a temporary bridge between the old top-level functions
// of EvaluationNode and their eventual inclusion into the Coefficient class.
// Everywhere that used to call an EvaluationNode in the C code now refers
// to this instead. Once complete and containing original Coefficient
// class stuff rename this class Coefficient.

class CoefficientNode {
  public:
    typedef shared_ptr<CoefficientNode> Ptr;
    typedef std::vector<CoefficientNode::Ptr> ptr_vector;

    CoefficientNode(const Coefficient & coeff) : coeff_(coeff) {
      preferred_method_ = 0;
    }

    void add_EvalNode(int method, const EvaluationNode::Ptr node);
    const Coefficient & coeff() const;

    Poly3T<double> value();
    template <typename fT> Poly3T<fT> valueT();

  private:
    int preferred_method_;
    std::map<int, EvaluationNode::Ptr> nodes_;
    const Coefficient coeff_;
};

/**********************************************************************
 *                                                                    *
 * EvaluationList                                                     *
 *                                                                    *
 **********************************************************************/

class EvaluationList {
  public:
    struct  node_cmp {
      bool operator()(const EvaluationNode::Ptr & node1,
                      const EvaluationNode::Ptr & node2);
      };

    // void set_evalution_method(int method); (deprecated)
    int get_method() const;

    typedef std::list<EvaluationNode::Ptr> Container;
//        typedef std::pair<Container::iterator, bool> ContainerIns;

    explicit EvaluationList(CoefficientPool & pool, int method,
                            bool is_primary = false);

    EvaluationNode::Ptr find(const Coefficient & coeff);
    EvaluationNode::Ptr add(CoefficientNode & cnode);
    EvaluationNode::Ptr add(const Coefficient & coeff);

    void clear_values();
    bool evaluate();
    bool is_primary() const;

    void print(bool print_dependencies = true) const;
    void set_fail(); // sets success_ to false

  private:
    EvaluationList::Container coefficients_;
    bool is_primary_;
    CoefficientPool & pool_;
    const int method_;
    // gets set to false if the evaluation runs into an unacceptable error
    bool success_;
};

/**********************************************************************
 *                                                                    *
 * CoefficientPool                                                    *
 *                                                                    *
 **********************************************************************/

class CoefficientPool {
  public:
    typedef std::vector<int> IntVec;
    typedef std::vector<std::string> StringVec;

    explicit CoefficientPool(ScalarIntInterface_t* scalar_if,
                             TensorIntInterface_t* tensor_if,
                             int method)
    : scalar_if_(scalar_if), tensor_if_(tensor_if), method_(method) {}
    ~CoefficientPool();

    void add_method(int method);
    CoefficientNode::Ptr create_coefficient(int points, InvariantGroup inv,
                                            const IntVec & indices,
                                            bool isRescue = false);

    CoefficientNode::Ptr create_coefficient(KinematicsMap *kmap,
                                            int points,
                                            const std::vector<IntVec> & momenta,
                                            const IntVec & masses,
                                            const IntVec & indices);

    CoefficientNode::Ptr create_coefficient(KinematicsMap *kmap,
                                            int points,
                                            const StringVec & momenta,
                                            const StringVec & masses,
                                            const IntVec & indices);

    CoefficientNode::Ptr add_coefficient(const Coefficient & coeff,
                                         bool isRescue);

    // adds a CoefficientNode to the pool without any other effects
    CoefficientNode::Ptr add_CoeffNode(const Coefficient & coeff);

    void clear_values();
    bool complete();

    bool operator==(const Coefficient & coeff);

    // returns a vector of all allowed methods for this many points
    IntVec get_all_methods(int points) const;
    int get_method() const;
    // returns highest-priority method appropriate for the number of points
    int get_primary_method(int points) const;

    void print();

    ScalarIntInterface_t* scalar_interface() const;
    const TensorIntInterface_t* tensor_interface() const;

  private:
    // Points to each method-dependent EvalList
    typedef std::map<int, EvaluationList> MethodTree;
    // InvSubTree is a map between hashes of invariants and the
    // appropriate lists of coefficients to evaluate (method-dependent)
    typedef std::map<int, CoefficientNode::Ptr> CoeffMap; // first is index hash
    struct CoefficientGroup {
      MethodTree m_tree;
      CoeffMap coeffs; // consider changing to unordered_map
    };
    typedef std::map<Invariants<fT1>::Hash_T, CoefficientGroup> InvSubTree;
    // Basic storage is a map between number of points of int and InvSubTree
    typedef std::map<int, InvSubTree> Tree;

    void add_alt_nodes(int points, std::vector<int> & alt_methods,
                       CoefficientNode::Ptr & p_cnode,
                       Invariants<fT1>::Hash_T invhash);
    CoefficientNode::Ptr add_CoeffNode(const Coefficient & coeff,
                                       bool & is_new);
    CoefficientNode::Ptr new_invariant(InvSubTree & subtree,
                                       Invariants<fT1>::Hash_T invhash,
                                       const Coefficient & coeff);
    Tree pool_;
    ScalarIntInterface_t* scalar_if_;
    TensorIntInterface_t* tensor_if_;
    int method_;
};

/**********************************************************************
 *                                                                    *
 * Inlined functions                                                  *
 *                                                                    *
 **********************************************************************/

inline Poly3T<double> CoefficientNode::value() {
  // Note: we assume preferred_method_ exists!
  // Should be safe with current implementation, but be careful if rewriting
  if (nodes_[preferred_method_]->is_cayley_zero()) {
    nodes_.erase(preferred_method_);
    preferred_method_ = PV_Reduction;
  }
  return nodes_[preferred_method_]->value();
}

inline const Coefficient & CoefficientNode::coeff() const {
  return coeff_;
}

template <typename fT> inline Poly3T<fT> CoefficientNode::valueT() {
  if (nodes_[preferred_method_]->is_cayley_zero()) {
    nodes_.erase(preferred_method_);
    preferred_method_ = PV_Reduction;
  }
  return nodes_[preferred_method_]->valueT<fT>();
}

inline void CoefficientPool::add_method(int method) {
  method_ |= method;
}

inline int CoefficientPool::get_method() const {
  return method_;
}

inline ScalarIntInterface_t* CoefficientPool::scalar_interface() const {
  return scalar_if_;
}

inline const TensorIntInterface_t* CoefficientPool::tensor_interface() const {
  return tensor_if_;
}

inline const Coefficient & EvaluationNode::coeff() const {
  return coeff_;
}

inline bool EvaluationNode::is_cayley_zero() {
  if (evaluation_level_ < 1) {
    try {
      evaluate();
    }
  catch (CayleyZeroException & cze) { // This only happens for DD nodes
#ifdef DEBUG
    std::cout << "Found zero Cayley determinant: "
              << cze.val_ << std::endl;
#endif //DEBUG
    is_cayley_zero_ = true;
    }
  }
  return is_cayley_zero_;
}

template <typename fT> inline Poly3T<fT> EvaluationNode::valueT() {
  not_eval_msg();
  std::cerr << "This precision level is not implemented." << std::endl;
  return Poly3T<fT>(0.,0.,0.,-2);
}

template <> inline Poly3T<fT1> EvaluationNode::valueT<fT1>() {
  if (evaluation_level_ >= 1) {
    return value1_;
  } else {
    evaluate();

    return valueT<fT1>();
  }
}
#if MAX_PREC_LVL >=2
template <> inline Poly3T<fT2> EvaluationNode::valueT<fT2>() {
  if (evaluation_level_ >= 2) {
    return value2_;
  } else {
    evaluate();
    return valueT<fT2>(); // infinite recursion if something goes wrong...
  }
}
#endif
#if MAX_PREC_LVL >=3
template <> inline Poly3T<fT3> EvaluationNode::valueT<fT3>() {
  if (evaluation_level_ >= 3) {
    return value3_;
  } else {
    evaluate();
    return valueT<fT3>(); // infinite recursion if something goes wrong...
  }
}
#endif

inline Poly3T<double> EvaluationNode::value() {
  return valueT<double>();
}

inline void EvaluationNode::not_eval_msg() const {
  std::cerr << "ERROR: value for coefficient not yet evaluated" << std::endl;
  std::cerr << "Coefficient: " << std::endl;
  coeff_.print(std::cout);
  exit(1);
}

inline int EvaluationList::get_method() const {
  return method_;
}

inline bool EvaluationList::is_primary() const {
  return is_primary_;
}

inline void EvaluationList::set_fail() {
  success_ = false;
}

// non-inlined template functions

template <typename fT>
Poly3T<fT> ContractedFirstProp::value() const {
  if (_deps.begin()->first <= 0) {
    assert(_deps.size() <= 1);
    std::vector<std::pair<int, std::vector<CoefficientNode::Ptr> > >::const_iterator i = _deps.begin();
    return NumUtil::signpow(i->first)*i->second.begin()->get()->valueT<fT>();
  } else {
    Poly3T<fT> result(0., 0., 0., -2);

    for (std::vector<std::pair<int, CoefficientNode::ptr_vector> >::const_iterator i = _deps.begin();
         i != _deps.end();  ++i) {
      fT binomial = NumUtil::binomial<fT>(_deps.size() - 1, i->first-1);
      for (std::vector<CoefficientNode::Ptr>::const_iterator j = i->second.begin();
           j != i->second.end(); ++j) {
        result += (j->get()->valueT<fT>())*binomial;
      }
    }
    return NumUtil::signpow(_deps.size()-1)*result;
  }
}

#endif  // EVALUATION_H_
