#ifndef FOUR_VECTOR_H_
#define FOUR_VECTOR_H_

#include <iostream>
#include <sstream>
#include <string>

/**********************************************************************
 *                                                                    *
 * FourVectorT                                                         *
 *                                                                    *
 * Simple implementation of Minkowski vectors, including basic        *
 * algebra and publicly accessible members.                           *
 *                                                                    *
 **********************************************************************/

template <typename fT> class FourVectorT {
  public:
    FourVectorT() : p0(0), p1(0), p2(0), p3(0) {}

    FourVectorT(const FourVectorT & fv)
    : p0(fv.p0), p1(fv.p1), p2(fv.p2), p3(fv.p3) {}

    template <typename T2> FourVectorT<fT>(const FourVectorT<T2> & fv) {
      p0 = fT(fv.p0);
      p1 = fT(fv.p1);
      p2 = fT(fv.p2);
      p3 = fT(fv.p3);
    }

    FourVectorT(fT x0, fT x1, fT x2, fT x3)
    : p0(x0), p1(x1), p2(x2), p3(x3) {}

    void boost(int dir, fT gamma, fT gammabeta);
    void rotate_xy(fT vz, fT sint, fT cost, fT sinp, fT cosp);

    // operator overloading for standard algebra
    FourVectorT<fT>  operator+(const FourVectorT<fT> & fv);
    FourVectorT<fT>& operator+=(const FourVectorT<fT> & fv);
    FourVectorT<fT>  operator-(const FourVectorT<fT> & fv);
    FourVectorT<fT>& operator-=(const FourVectorT<fT> & fv);

    // Negation operator
    FourVectorT<fT> operator-() const;

    FourVectorT<fT> negate(); // Redundant
    FourVectorT<fT> negateSpatial();

    fT comp(int i);
    fT abs();

    // Computes the scalar product between two FourVectorTs.
    // ( signature g_mu,nu = (+1,-1,-1,-1) )
    // TODO: put outside class?
    static fT scalar_product(const FourVectorT<fT> & v1,
                             const FourVectorT<fT> & v2);

    // Computes the contraction of four FourVectorTs with the
    // totally antisymmetric Levi-Civita tensor e_(mu,nu,ro.sig)
    static fT antisym_product(const FourVectorT<fT> & v1,
                              const FourVectorT<fT> & v2,
		              const FourVectorT<fT> & v3,
		              const FourVectorT<fT> & v4);

    void print(std::ostream & strm = std::cout, std::string prefix = "");
    void print(int prec, int width, std::ostream & strm = std::cout,
               std::string prefix = "");

    //std::string fortran90_string(std::string suffix = "_dp",
    std::string recolaString(std::string suffix = "_dp",
                                 bool split = false, std::string padding = "");

    // public, we know what we do
    fT p0, p1, p2, p3;
};

typedef FourVectorT<double> FourVector;

// Template functions

template <typename fT> inline FourVectorT<fT> FourVectorT<fT>::operator+(
    const FourVectorT<fT> & fv) {
  return FourVectorT<fT>(p0 +fv.p0, p1 + fv.p1, p2 + fv.p2, p3 + fv.p3);
}

template <typename fT> inline FourVectorT<fT>& FourVectorT<fT>::operator+=(
    const FourVectorT<fT> & fv) {
  p0 += fv.p0;
  p1 += fv.p1;
  p2 += fv.p2;
  p3 += fv.p3;
  return *this;
}

template <typename fT> inline FourVectorT<fT> FourVectorT<fT>::operator-(
    const FourVectorT<fT> & fv) {
  return FourVectorT<fT>(p0 - fv.p0, p1 - fv.p1, p2 - fv.p2, p3 - fv.p3);
}

template <typename fT> inline FourVectorT<fT>& FourVectorT<fT>::operator-=(
    const FourVectorT<fT> & fv) {
  p0 -= fv.p0;
  p1 -= fv.p1;
  p2 -= fv.p2;
  p3 -= fv.p3;
  return *this;
}

template <typename fT> inline FourVectorT<fT>
FourVectorT<fT>::operator-() const {
  return FourVectorT<fT>(-p0, -p1, -p2, -p3);
}

template <typename fT> inline FourVectorT<fT> operator*(
    const FourVectorT<fT> & vec, const fT & val) {
  return FourVectorT<fT>(val*vec.p0, val*vec.p1, val*vec.p2, val*vec.p3);
}

template <typename fT> inline FourVectorT<fT> FourVectorT<fT>::negate() {
  return FourVectorT<fT>(-p0, -p1, -p2, -p3);
}

template <typename fT> inline FourVectorT<fT>
FourVectorT<fT>::negateSpatial() {
  return FourVectorT<fT>(p0, -p1, -p2, -p3);
}

template <typename fT> inline fT FourVectorT<fT>::abs() {
  return p0*p0 - p1*p1 - p2*p2 - p3*p3;
}

template <typename fT> inline fT FourVectorT<fT>::comp(int i) {
  fT retval = 0;
  switch (i) {
    case 0: retval = p0;
      break;
    case 1: retval = p1;
      break;
    case 2: retval = p2;
      break;
    case 3: retval = p3;
      break;
    default: std::cerr << "error: FourVectorT::comp: index " << i
                       << " ne [0,3]" << std::endl;
      exit(1);
    }
  return retval;
}

template <typename fT> inline fT FourVectorT<fT>::scalar_product(
    const FourVectorT & v1, const FourVectorT & v2) {
  return v1.p0*v2.p0 - v1.p1*v2.p1 - v1.p2*v2.p2 - v1.p3*v2.p3;
}

// non-inlined functions

template <typename fT> fT FourVectorT<fT>::antisym_product(
    const FourVectorT & v1, const FourVectorT & v2,
    const FourVectorT & v3, const FourVectorT & v4) {
  return + v1.p0*v2.p1 * (v3.p2*v4.p3 - v3.p3*v4.p2)
         + v1.p0*v2.p2 * (v3.p3*v4.p1 - v3.p1*v4.p3)
         + v1.p0*v2.p3 * (v3.p1*v4.p2 - v3.p2*v4.p1)
         + v1.p1*v2.p0 * (v3.p3*v4.p2 - v3.p2*v4.p3)
         + v1.p1*v2.p2 * (v3.p0*v4.p3 - v3.p3*v4.p0)
         + v1.p1*v2.p3 * (v3.p2*v4.p0 - v3.p0*v4.p2)
         + v1.p2*v2.p0 * (v3.p1*v4.p3 - v3.p3*v4.p1)
         + v1.p2*v2.p1 * (v3.p3*v4.p0 - v3.p0*v4.p3)
         + v1.p2*v2.p3 * (v3.p0*v4.p1 - v3.p1*v4.p0)
         + v1.p3*v2.p0 * (v3.p2*v4.p1 - v3.p1*v4.p2)
         + v1.p3*v2.p1 * (v3.p0*v4.p2 - v3.p2*v4.p0)
         + v1.p3*v2.p2 * (v3.p1*v4.p0 - v3.p0*v4.p1);
}

template <typename fT> void FourVectorT<fT>::boost(int dir,
                                                   fT gamma, fT gammabeta) {
  fT pp0, pp1, pp2, pp3;
  switch(dir) {
    case 1: pp0 = gamma*p0 + gammabeta*p1;
            pp1 = gammabeta*p0 + gamma*p1;
            pp2 = p2;
            pp3 = p3;
            break;
    case 2: pp0 = gamma*p0 + gammabeta*p2;
            pp1 = p2;
            pp2 = gammabeta*p0 + gamma*p2;
            pp3 = p3;
            break;
    case 3: pp0 = gamma*p0 + gammabeta*p3;
            pp1 = p1;
            pp2 = p2;
            pp3 = gammabeta*p0 + gamma*p3;
            break;
  }
  p0 = pp0;
  p1 = pp1;
  p2 = pp2;
  p3 = pp3;
}

template <typename fT> void FourVectorT<fT>::rotate_xy(fT vz,
                                                       fT sint, fT cost,
                                                       fT sinp, fT cosp) {
  fT pp1, pp2, pp3;
  pp1 = vz*(cosp*(cost*p1 + sint*p3) - sinp*p2);
  pp2 = vz*(sinp*(cost*p1 + sint*p3) + cosp*p2);
  pp3 = vz*(-sint*p1 + cost*p3);

  p1 = pp1;
  p2 = pp2;
  p3 = pp3;
}

template <typename fT> void FourVectorT<fT>::print(std::ostream & strm,
                                                   std::string prefix) {
  strm << prefix << p0 << " " << p1 << " " << p2 << " " << p3 << std::endl;
}

template <typename fT> void FourVectorT<fT>::print(int prec, int width,
                                                   std::ostream & strm,
                                                   std::string prefix) {
  strm << std::setprecision(prec)
       << prefix
       << std::setw(width) << p0 << "\t"
       << std::setw(width) << p1 << "\t"
       << std::setw(width) << p2 << "\t"
       << std::setw(width) << p3 << std::endl;
}

/*  fortran90_string returns a string containing the FourVector in a format
    that can easily be copied into Fortran codes.
    suffix: Appended to every number (e.g. d0, _dp)
    split: If true, the y- and z- components will be printed on a new line.
    padding: The space printed before components on a new line.
 */
//template <typename fT> std::string FourVectorT<fT>::fortran90_string(
template <typename fT> std::string FourVectorT<fT>::recolaString(
    std::string suffix, bool split, std::string padding) {
  std::stringstream stream;

  stream << std::setprecision(16);
  stream << "[ ";
  stream << p0 << suffix << ", ";
  stream << p1 << suffix << ", ";

  if (split) stream << "&\n" << padding;

  stream << p2 << suffix << ", ";
  stream << p3 << suffix << " ]";

  return stream.str();
}

#endif // FOUR_VECTOR_H_
