#ifndef INTERFACES_H_
#define INTERFACES_H_

#include <string>
#include <iostream>
#include <cstdlib>
#include <map>
#include <algorithm> // min
#include <complex>

#ifdef QCDLOOP
#include "cqcdloop.h" // QCDLoop
#endif

#ifdef QCDLOOP2
#include "qcdloop/qcdloop.h" // QCDLoop2
#endif

#ifdef LOOPTOOLS
#include "clooptools.h" // LoopTools
#endif

#ifdef ONELOOP
// OneLoop's C++ interface header formely assumed namespace std.
// TODO: We have replaced OL's header with our own version, so we should
// correct other parts of the code that assume namespace std.
using namespace std;
#include "cavh_olo.h" // OneLoop
#endif

#ifdef COLLIER
#include "ccollier.h" // Collier
#endif

#include "common.h"
#include "numerics.h"
#include "kinematics.h"
#include "tcomplex.h"

namespace Interfaces {
  template <typename fT> void if_c_kinematics(
      const typename Invariants<fT>::Ptr & inv,
      typename Vector<fT>::type & retval);
  template <typename fT> void if_d_kinematics(
      const typename Invariants<fT>::Ptr & inv,
      typename Vector<fT>::type & retval);
  template <typename fT> void if_e_kinematics(
      const typename Invariants<fT>::Ptr & inv,
      typename Vector<fT>::type & retval);
  template <typename fT> void if_arguments(
      const typename Invariants<fT>::Ptr & inv,
      typename Vector<fT>::type & retval);
};

// IntInterface (virtual)

class IntInterface {
  public:
    IntInterface(int relprec);
    template <typename fT> void adjust_kinematics(
        fT* inv,
        typename tcomplex<fT>::type *masses,
        int invlen, int mlen) const;
    double get_musq() const;
    virtual void set_musq(double musq);

    const int relprec_;
    const double true_relprec_; // true_relprec_ = 10^{-relprec_}

  protected:
#ifdef DEBUG
//    static int callCount_ = 0;
    static int callCount_;
#endif
    double musq_;
};

/**********************************************************************
 *                                                                    *
 * ScalarIntInterface (virtual)                                       *
 *                                                                    *
 **********************************************************************/

// Note on the templating of scalar interfaces: functions internal to tred
// use tcomplex<fT>::type so that non-standard complex types can be used.
// For simplicity of interface, outward-facing functions such as A0 can only
// use C++ std::complex at present; otherwise their callers would need to know
// how to use tcomplex<fT>::type (e.g., processconst for EW counterterms).

class ScalarIntInterface : virtual public IntInterface {
  public:
    ScalarIntInterface(int relprec);

    template <typename fT> Poly3T<fT> A0(std::complex<fT> m0);
    template <typename fT> Poly3T<fT> B0(fT psq, std::complex<fT> m0,
                                         std::complex<fT> m1);
    template <typename fT> Poly3T<fT> evaluate_scalar(
        const typename Invariants<fT>::Ptr & inv);
};

/**********************************************************************
 *                                                                    *
 * QLInterface                                                        *
 *                                                                    *
 **********************************************************************/

#ifdef QCDLOOP
class QLInterface : public ScalarIntInterface {
  public:
    explicit QLInterface(int relprec);
    ~QLInterface();
    template <typename fT>
    Poly3T<fT> A0(std::complex<fT> m0);
    template <typename fT> Poly3T<fT> B0(fT psq, std::complex<fT> m0,
                                         std::complex<fT> m1);
    template <typename fT> Poly3T<fT> evaluate_scalar(
        const typename Invariants<fT>::Ptr & inv);

  private:
    static bool isInitialized_;
};
#endif

/**********************************************************************
 *                                                                    *
 * QL2Interface                                                       *
 *                                                                    *
 **********************************************************************/

#ifdef QCDLOOP2
class QL2Interface : public ScalarIntInterface {
  public:
    explicit QL2Interface(int relprec);
    ~QL2Interface();
    template <typename fT>
    Poly3T<fT> A0(std::complex<fT> m0);
    template <typename fT> Poly3T<fT> B0(fT psq, std::complex<fT> m0,
                                         std::complex<fT> m1);
    template <typename fT> Poly3T<fT> evaluate_scalar(
        const typename Invariants<fT>::Ptr & inv);

  private:
  /* Implementation note: QCDLoop2 can use a single object of type
     QCDLoop and automatically detect the topology, meaning we only
     really need one object here and the code would be cleaner.
     But it might be faster to use the explicit separate objects
     since we already know the topology by time we call the scalar
     integrals. Efficiency should be tested. */

    ql::TadPole<std::complex<double>, std::complex<double>, double> tp;
    ql::Bubble<std::complex<double>, std::complex<double>, double> bub;
    ql::Triangle<std::complex<double>, std::complex<double>, double> tri;
    ql::Box<std::complex<double>, std::complex<double>, double> box;

  // Quadruple precision

    ql::TadPole<ql::qcomplex, ql::qcomplex, ql::qdouble> tpq;
    ql::Bubble<ql::qcomplex, ql::qcomplex, ql::qdouble> bubq;
    ql::Triangle<ql::qcomplex, ql::qcomplex, ql::qdouble> triq;
    ql::Box<ql::qcomplex, ql::qcomplex, ql::qdouble> boxq;
};
#endif

// TensorIntInterface (virtual)

class TensorIntInterface : virtual public IntInterface {
  public:
    TensorIntInterface(int relprec);
    template <typename fT> Poly3T<fT> evaluate_tensor_coeff(
        const std::vector<int> & indices,
        const typename Invariants<fT>::Ptr & inv,
        bool & success) const;
};

// DerivativeIntInterface
// This class is for interfaces to scalar derivative integral libraries.

class DerivativeIntInterface : virtual public IntInterface {
  public:
    DerivativeIntInterface(int relprec);
    template <typename fT> Poly3T<fT> dB0(
        const typename Invariants<fT>::Ptr & inv);

    // alternate method that doesn't require knowledge of Invariants structure
    template <typename fT> Poly3T<fT> dB0(fT psq, std::complex<fT> m0,
                                          std::complex<fT> m1);
};

/**********************************************************************
 *                                                                    *
 * TRDerivativeInterface                                              *
 *                                                                    *
 **********************************************************************/

// This interface implements B derivatives directly without an external library

class TRDerivativeInterface : public DerivativeIntInterface {
  public:
    TRDerivativeInterface(int relprec);

    template <typename fT> Poly3T<fT> dB0(
        const typename Invariants<fT>::Ptr & inv);

    template <typename fT> Poly3T<fT> dB0(fT psq, std::complex<fT> m0,
                                          std::complex<fT> m1);

  private:
    // consider moving to type-dependent precision
    static const double epsFeyn;
};

/**********************************************************************
 *                                                                    *
 * LTInterface                                                        *
 *                                                                    *
 **********************************************************************/

#ifdef LOOPTOOLS
class LTInterface : public ScalarIntInterface, public TensorIntInterface,
                    public DerivativeIntInterface {
  public:
    LTInterface(int relprec);
    ~LTInterface();

    template <typename fT> Poly3T<fT> evaluate_scalar(
        const typename Invariants<fT>::Ptr & inv);
    template <typename fT> Poly3T<fT> A0(std::complex<fT> m0);
    template <typename fT> Poly3T<fT> B0(fT psq, std::complex<fT> m0,
                                         std::complex<fT> m1);

    template <typename fT> Poly3T<fT> evaluate_tensor_coeff(
        const std::vector<int> & indices,
        const typename Invariants<fT>::Ptr & inv,
        bool & success) const;

    template <typename fT> Poly3T<fT> dB0(
        const typename Invariants<fT>::Ptr & inv);

    template <typename fT> Poly3T<fT> dB0(fT psq, std::complex<fT> m0,
                                          std::complex<fT> m1);

    void set_musq(double musq);

  private:
    // maps between TRmgr's and LT's indices, obtained from LT/includes/defs.h

    static int TRLTb[][2];
    static int TRLTdb[][2];
    static int TRLTc[][2];
    static int TRLTd[][2];
    static int TRLTe[][2];

    static int TRLTb_length;
    static int TRLTdb_length;
    static int TRLTc_length;
    static int TRLTd_length;
    static int TRLTe_length;

    static std::map<int, int> LTmapB;
    static std::map<int, int> LTmapdB;
    static std::map<int, int> LTmapC;
    static std::map<int, int> LTmapD;
    static std::map<int, int> LTmapE;

    static int numInstances_;
};
#endif // LOOPTOOLS

#ifdef ONELOOP
class OLInterface : public ScalarIntInterface, public DerivativeIntInterface {
  public:
    OLInterface(int relprec);
    void set_musq(double musq);

    template <typename fT> Poly3T<fT> evaluate_scalar(
        const typename Invariants<fT>::Ptr & inv);
    template <typename fT> Poly3T<fT> A0(std::complex<fT> m0);
    template <typename fT> Poly3T<fT> B0(fT psq, std::complex<fT> m0,
                                         std::complex<fT> m1);
    template <typename fT> Poly3T<fT> dB0(
        const typename Invariants<fT>::Ptr & inv);
    template <typename fT> Poly3T<fT> dB0(fT psq, std::complex<fT> m0,
                                          std::complex<fT> m1);

  private:
    // this should be set far lower than tred's internal on shell threshold;
    // this is an absolute scale
    double onshell_thresh;
  
};
#endif

/**********************************************************************
 *                                                                    *
 * CollierInterface                                                   *
 *                                                                    *
 **********************************************************************/

#ifdef COLLIER
class CollierInterface : public ScalarIntInterface, public TensorIntInterface {

  public:
    CollierInterface(int relprec);
    template <typename fT> Poly3T<fT> evaluate_scalar(
        const typename Invariants<fT>::Ptr & inv);
    template <typename fT> Poly3T<fT> A0(std::complex<fT> m0);
    template <typename fT> Poly3T<fT> B0(fT psq, std::complex<fT> m0,
                                         std::complex<fT> m1);

    template <typename fT> Poly3T<fT> evaluate_tensor_coeff(
        const std::vector<int> & indices,
        const typename Invariants<fT>::Ptr & inv,
        bool & success) const;

    //template <typename fT> Poly3T<fT> dB0(
    //    const typename Invariants<fT>::Ptr & inv);

    //template <typename fT> Poly3T<fT> dB0(fT psq, std::complex<fT> m0,
    //                                      std::complex<fT> m1);

    void set_musq(double musq);

  private:
    static int numInstances_;
};
#endif // COLLIER

// Interface for internal pole library
// Right now, the internal library is available in double precision only. If
// there are ever checks that require varying degrees of tolerance, one might
// want to template the coefficient calls.

class PoleCheckInterface : public TensorIntInterface {
  public:
    PoleCheckInterface(int relprec);

    template <typename fT> Poly3T<fT> evaluate_tensor_coeff(
        const std::vector<int> & indices,
        const typename Invariants<fT>::Ptr & inv,
        bool & success) const;

  private:
    double c_coeff(const std::vector<int> & indices,
                   const Vector<double>::type & mominv,
                   const double masses[3],
                   bool & success) const;
    double d_coeff(const std::vector<int> & indices,
                   const Vector<double>::type & mominv,
                   const double masses[4],
                   bool & success) const;
    double e_coeff(const std::vector<int> & indices,
                   const Vector<double>::type & mominv,
                   const double masses[5],
                   bool & success) const;

    double C00() const;
    double C0000(const Vector<double>::type & mominv,
                 const double masses[3]) const;
    double C00i() const;
    double C00ii() const;
    double C00ij() const;

    double D00() const;
    double D0000() const;
    double D00i() const;
    double D00ij() const;
    double D0000i() const;
    double D00ijk() const;

    double E00() const;
    double E0000() const;
    double E000000() const;
};

// Template functions
// The default behavior for all scalar/tensor libraries is to use the highest
// precision available. Specific instantiations if there is more than one
// available are found in interfaces.cpp.

namespace Interfaces  {
  template <typename fT> void if_c_kinematics(
      const typename Invariants<fT>::Ptr & inv,
      typename Vector<fT>::type & retval) {
    retval.reserve(3);
    retval.push_back(inv->external_invariants(0, 0));
    retval.push_back(inv->external_invariants(1, 1));
    retval.push_back(  inv->external_invariants(0, 0) +
                     2*inv->external_invariants(0, 1) +
                       inv->external_invariants(1, 1));
  }

  template <typename fT> void if_d_kinematics(
      const typename Invariants<fT>::Ptr & inv,
      typename Vector<fT>::type & retval)  {
    retval.reserve(6);
    retval.push_back(inv->external_invariants(0, 0));
    retval.push_back(inv->external_invariants(1, 1));
    retval.push_back(inv->external_invariants(2, 2));
    retval.push_back(inv->external_invariants(3, 3));
    retval.push_back(inv->external_invariants(0, 0)
                     + 2*inv->external_invariants(0, 1)
                     + inv->external_invariants(1, 1));
    retval.push_back(inv->external_invariants(1, 1)
                     + 2*inv->external_invariants(1, 2)
                     + inv->external_invariants(2, 2));
  }

  template <typename fT> void if_e_kinematics(
      const typename Invariants<fT>::Ptr & inv,
      typename Vector<fT>::type & retval) {
    retval.reserve(10);
    // precalc redundant momentum
    retval.push_back(inv->external_invariants(0, 0));
    retval.push_back(inv->external_invariants(1, 1));
    retval.push_back(inv->external_invariants(2, 2));
    retval.push_back(inv->external_invariants(3, 3));
    retval.push_back(inv->external_invariants(4, 4));
    retval.push_back(inv->external_invariants(0, 0)
                     + 2*inv->external_invariants(0, 1)
                     + inv->external_invariants(1, 1));
    retval.push_back(inv->external_invariants(1, 1)
                     + 2*inv->external_invariants(1, 2)
                     + inv->external_invariants(2, 2));
    retval.push_back(inv->external_invariants(2, 2)
                     + 2*inv->external_invariants(2, 3)
                     + inv->external_invariants(3, 3));
    retval.push_back(inv->external_invariants(3, 3)
                     + 2*inv->external_invariants(3, 4)
                     + inv->external_invariants(4, 4));
    retval.push_back(inv->external_invariants(0, 0)
                     + 2*inv->external_invariants(0, 4)
                     + inv->external_invariants(4, 4));
}

  template <typename fT> void if_arguments(
      const typename Invariants<fT>::Ptr & inv,
      typename Vector<fT>::type & retval) {
    switch (inv->points()) {
      case 2: retval = typename Vector<fT>::type(1, inv->external_invariants(0, 0));
              break;
      case 3: if_c_kinematics<fT>(inv, retval);
              break;
      case 4: if_d_kinematics<fT>(inv, retval);
              break;
      case 5: if_e_kinematics<fT>(inv, retval);
              break;
    }
  }
}  // Interfaces

// This function is for checking whether invariants/masses are "close enough"
// to be identical. This has to be tuned somewhat and will depend on the PSP
// generator. Ideally we would pass some sort of flag through tred and/or
// keep track of invariant masses, but at present the latter is unfeasible as
// it would require leaving the last momentum without a momentum conservation
// substitution.

template <typename fT> void IntInterface::adjust_kinematics(
    fT* inv,
    typename tcomplex<fT>::type *masses,
    int invlen, int mlen) const {
  for (int i = 0; i < invlen; ++i) {
    for (int j = 0; j < mlen; ++j) {
      fT realmass = real(masses[j]);
      if (NumUtil::equal_float<fT>(inv[i], realmass, true_relprec_, 1e-30))
        inv[i] = realmass;
    }
    for (int j = 0; j < invlen; ++j) {
      if (NumUtil::zero_float<fT>(inv[i], inv[j], true_relprec_))
        inv[i] = fT(0);
    }
  }
}

template <typename fT> Poly3T<fT> DerivativeIntInterface::dB0(
    const typename Invariants<fT>::Ptr & inv) {

  if (inv->points() != 2) {
    std::cerr << "Error, derivatives for anything other than B functions unsupported."
              << std::endl;
    exit(0);
  }

  typename Vector<fT>::type args;
  Interfaces::if_arguments<fT>(inv, args);

  // This section will need to be rewritten if we ever start calling derivatives
  // for anything other than B's
  fT psq = args[0];
  std::complex<fT> m0 = inv->mass(0);
  std::complex<fT> m1 = inv->mass(1);

  return dB0(psq, m0, m1);
}

#ifdef QCDLOOP
template <typename fT> Poly3T<fT> QLInterface::A0(std::complex<fT> m0) {
  double realres[3], imagres[3];
  double musqtmp = musq_;
  double masstr[1] = {real(m0)};

#ifdef DEBUG
  callCount_++;
#endif

  if (imag(m0) > relprec_*masstr[0])
    std::cout << "WARNING: attempt to evaluate A function with complex"
              << "mass in QCDLoop!" << std::endl
              << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << std::endl;

  FORTSUB(qlsi1)(masstr, &musqtmp, realres, imagres);

  return
    Poly3T<fT>(std::complex<fT>(realres[0], imagres[0]),
               std::complex<fT>(realres[1], imagres[1]),
               std::complex<fT>(realres[2], imagres[2]), -2);
}

template <typename fT> Poly3T<fT> QLInterface::B0(fT psq,
                                                  std::complex<fT> m0,
                                                  std::complex<fT> m1) {
  double realres[3], imagres[3];
  double musqtmp = musq_;
  double m0r = real(m0);
  double m1r = real(m1);
  double masses[2] = {m0r, m1r};
  double invtr[1] = {psq};

#ifdef DEBUG
  callCount_++;
#endif

  if ((imag(m0) > m0r*relprec_) || (imag(m1) > m1r*relprec_))
    std::cout << "WARNING: attempt to evaluate B function with complex"
              << "mass in QCDLoop!" << std::endl
              << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << std::endl;

  FORTSUB(qlsi2)(&(invtr[0]), &(masses[0]), &(masses[1]), &musqtmp,
                 realres, imagres);

  return
    Poly3T<fT>(std::complex<fT>(realres[0], imagres[0]),
               std::complex<fT>(realres[1], imagres[1]),
               std::complex<fT>(realres[2], imagres[2]), -2);
}

template <typename fT> Poly3T<fT> QLInterface::evaluate_scalar(
    const typename Invariants<fT>::Ptr & inv) {
  typename Vector<fT>::type args;

  double realres[3], imagres[3];
  double musqtmp = musq_;
  double realmasses[4];
  std::complex<double> masses[4];
  double invtr[6];
  int points = inv->points();
  int numInv = (points*points - points)/2;

  Interfaces::if_arguments<fT>(inv, args);

#ifdef DEBUG
  callCount_++;
#endif

  for (int i = 0; i < inv->points(); ++i) {
    // QCDLoop 1 can only handle real masses; will produce wrong result if
    // complex!!
    realmasses[i] = real(inv->mass(i));
    masses[i] = realmasses[i];
  }

  // truncate values for specified precision, so that QL knows when
  // to switch to IR divergent evaluation

  for (int i = 0; i < numInv; ++i)
    invtr[i] = NumUtil::truncate<double>(args[i], musq_*true_relprec_);

  adjust_kinematics(invtr, masses, numInv, points);

  switch (inv->points()) {
    case 1: FORTSUB(qlsi1)(&(realmasses[0]), &musqtmp, realres, imagres);
            break;

    case 2: FORTSUB(qlsi2)(&(invtr[0]),
                           &(realmasses[0]), &(realmasses[1]), &musqtmp,
                           realres, imagres);
            break;

    case 3: FORTSUB(qlsi3)(&(invtr[0]), &(invtr[1]), &(invtr[2]),
                           &(realmasses[0]), &(realmasses[1]),
                           &(realmasses[2]), &musqtmp,
                           realres, imagres);
            break;

    case 4: FORTSUB(qlsi4)(&(invtr[0]), &(invtr[1]), &(invtr[2]),
                           &(invtr[3]), &(invtr[4]), &(invtr[5]),
                           &(realmasses[0]), &(realmasses[1]),
                           &(realmasses[2]), &(realmasses[3]), &musqtmp,
                           realres, imagres);
            break;
  }

  return
    Poly3T<fT>(std::complex<fT>(realres[0], imagres[0]),
               std::complex<fT>(realres[1], imagres[1]),
               std::complex<fT>(realres[2], imagres[2]), -2);
}
#endif // QCDLOOP

#ifdef QCDLOOP2
template <typename fT> Poly3T<fT> QL2Interface::A0(std::complex<fT> m0) {
  std::vector<ql::qcomplex> res(3);
  std::vector<ql::qcomplex> massvec;
  std::vector<ql::qdouble> invtrvec;

  massvec.push_back(m0);

#ifdef DEBUG
  callCount_++;
#endif

  tpq.integral(res, musq_, massvec, invtrvec);

  return Poly3T<fT>(res[2], res[1], res[0], -2);
}

template <typename fT> Poly3T<fT> QL2Interface::B0(fT psq,
                                                   std::complex<fT> m0,
                                                   std::complex<fT> m1) {
  typedef typename tcomplex<fT>::type cfT;

  std::vector<ql::qcomplex> res(3);
  cfT masses[2] = {m0, m1};
  fT invtr[1] = {psq};
  std::vector<ql::qcomplex> massvec;

#ifdef DEBUG
  callCount_++;
#endif

  for (int i = 0; i < 2; ++i) {
    massvec.push_back(masses[i]);
  }
  std::vector<ql::qdouble> invtrvec(invtr, invtr+1);

  bubq.integral(res, musq_, massvec, invtrvec);

  return Poly3T<fT>(res[2], res[1], res[0], -2);
}

template <typename fT> Poly3T<fT> QL2Interface::evaluate_scalar(
    const typename Invariants<fT>::Ptr & inv) {
  typename Vector<fT>::type args;
  std::vector<ql::qcomplex> res(3);
  std::vector<ql::qcomplex> massvec;
  typename tcomplex<fT>::type masses[4];
  fT invtr[6];
  int points = inv->points();
  int numInv = (points*points - points)/2;

  Interfaces::if_arguments<fT>(inv, args);

#ifdef DEBUG
  callCount_++;
#endif

  for (int i = 0; i < points; ++i) {
    masses[i] = inv->mass(i);
  }

  // truncate values for specified precision, so that QL knows when
  // to switch to IR divergent evaluation

  for (int i = 0; i < numInv; ++i)
    invtr[i] = NumUtil::truncate<fT>(args[i], musq_*true_relprec_);

  adjust_kinematics(invtr, masses, numInv, points);

  for (int i = 0; i < points; ++i) {
    massvec.push_back({real(inv->mass(i)), imag(inv->mass(i))});
  }

  std::vector<ql::qdouble> invtrvec(invtr, invtr+numInv);

  switch (inv->points()) {
    case 1: tpq.integral(res, musq_, massvec, invtrvec);
            break;

    case 2: bubq.integral(res, musq_, massvec, invtrvec);
            break;

    case 3: triq.integral(res, musq_, massvec, invtrvec);
            break;

    case 4: boxq.integral(res, musq_, massvec, invtrvec);
            break;
  }

  return Poly3T<fT>(res[2], res[1], res[0], -2);
}

// Declare these specializations so we know they can be seen from the outside
template <> Poly3T<double> QL2Interface::A0(std::complex<double> m0);

template <> Poly3T<double> QL2Interface::B0(double psq,
                                            std::complex<double> m0,
                                            std::complex<double> m1);

template <> Poly3T<double> QL2Interface::evaluate_scalar(
    const typename Invariants<double>::Ptr & inv);
#endif // #ifdef QCDLOOP2

template <typename fT> Poly3T<fT> TRDerivativeInterface::dB0(
    fT psq, std::complex<fT> m0sq, std::complex<fT> m1sq) {
  typedef std::complex<fT> cfT;

  static const cfT cOne(1,0);
  static const cfT ieps(0.,epsFeyn);

  fT invtr[1] = {psq};
  cfT pole(0);
  cfT fin(0); // default is zero
  cfT minmass, maxmass;

  if (real(m0sq) < real(m1sq)) {
    minmass = m0sq;
    maxmass = m1sq;
  }
  else {
    minmass = m1sq;
    maxmass = m0sq;
  }

  if (fabs(psq) <= 0.) { // psq == 0
    if (real(minmass) <= 0. && real(maxmass) > 0.) { // one is positive, one zero
      fin = 0.5/maxmass;
    }
    else if (m0sq == m1sq && real(m0sq) > 0.) { // equal and positive
      fin = 1./(6.*maxmass);

    }
    else if (real(minmass) > 0.) { // unequal and positive
      // check below for which order is more numerically stable
      cfT massdiff = maxmass - minmass;
      cfT prefac = minmass*maxmass/(massdiff*massdiff*massdiff);
      fin = 0.5*(maxmass*maxmass-minmass*minmass)/(massdiff*massdiff*massdiff)
          - prefac*log(maxmass/minmass);
    } // else = 0
  }
  else { // psq != 0
    fT psqinv = 1/psq;
    if (real(maxmass) <= 0.) {
      fin = -psqinv;
    }
    else if (real(minmass) <= 0.) {
      // Change to first case: p^2 != M^2 if M^2 complex!
      if (real(maxmass) != invtr[0] || imag(maxmass) != 0.) {
        cfT logarg((maxmass-psq)/maxmass - ieps);
        fin = -psqinv - maxmass*psqinv*psqinv*log(logarg);
      }
      else {
        cfT invmass = 1./maxmass;
        pole = -0.5*invmass;
        fin = -invmass*(cOne + 0.5*log(musq_*invmass));
      }
    }
    else {
      cfT m0 = sqrt(maxmass);
      cfT m1 = sqrt(minmass);

      cfT rsum((maxmass + minmass - psq)/m0/m1 - ieps);
      fT rdiffSign = 1.;
      // choose sign of rdiff to be same as rsum for numerical stability
      // when calculating r
      if (real(rsum) < 0.) rdiffSign = -1.;
      cfT rdiff = rdiffSign*sqrt(rsum*rsum - cfT(4));
      cfT logr = log(0.5*(rsum+rdiff));

      cfT res = (minmass-maxmass)*psqinv*log(m1/m0)
              - m0*m1*psqinv*rdiff*logr
              - (cOne + rsum/rdiff*logr);
      fin = res*psqinv;
    }
  }

  return Poly3T<fT>(0., pole, fin, -2);
}

#ifdef LOOPTOOLS
template <typename fT> Poly3T<fT> LTInterface::evaluate_tensor_coeff(
    const std::vector<int> & indices,
    const typename Invariants<fT>::Ptr & inv,
    bool & success) const {
  typedef std::complex<double> cmplx;
  int64_t callindex;
  typename Vector<fT>::type mominv;
  cmplx masses[5];
  double invtr[10];
  int points = inv->points();
  int numInv = (points*points - points)/2;
  cmplx exp[3];

  Interfaces::if_arguments<fT>(inv, mominv);

  for (int i = 0; i < inv->points(); ++i) {
    masses[i] = cmplx(real(inv->mass(i)), imag(inv->mass(i)));
  }

  // truncate values for specified precision, so that QL knows when
  // to switch to IR divergent evaluation

  for (int i = 0; i < numInv; ++i) {
    invtr[i] = NumUtil::truncate<double>(mominv[i], musq_*true_relprec_);
  }

  adjust_kinematics(invtr, masses, numInv, points);

  // construct indices for tensor coeff evaluation

  int rep = 0;
  int powten = 1;

  for (int i = 0; i < indices.size(); ++i) {
    rep += powten*(indices[indices.size()-i-1]);
    powten *= 10;
  }
  rep += 9*powten; // leading 9 to know how many leading zeroes

  for (int i = 0; i <= 2; ++i) {
    setlambda(-i);
    int mapVal;
    cmplx result = cmplx(0, 0);

  switch (inv->points()) {
    case 2:
      mapVal = LTmapB.at(rep);
      result = B0iC(mapVal, invtr[0], masses[0], masses[1]);
      success = true;
      break;
    case 3:
      mapVal = LTmapC.at(rep);
      result = C0iC(mapVal, invtr[0], invtr[1], invtr[2],
                    masses[0], masses[1], masses[2]);
      success = true;
      break;
    case 4:
      mapVal = LTmapD.at(rep);
      result = D0iC(mapVal, invtr[0], invtr[1], invtr[2], invtr[3],
                    invtr[4], invtr[5],
                    masses[0], masses[1], masses[2], masses[3]);
      success = true;
      break;
    case 5:
/*      mapVal = LTmapE.at(rep);
      result = E0iC(mapVal, invtr[0], invtr[1], invtr[2], invtr[3],
                    invtr[4], invtr[5], invtr[6], invtr[7], invtr[8],
                    invtr[9], masses[0], masses[1], masses[2], masses[3],
                    masses[4]);
      success = true;*/
      success = false; // disabled due to probable LT bugs in E functions
      break;
    default:
      success = false;
      break;
    }
  exp[i] += result;
  }

  return Poly3T<fT>(exp[2],exp[1],exp[0],-2);
}

template <typename fT> Poly3T<fT> LTInterface::A0(std::complex<fT> m0) {
  typedef std::complex<double> cmplx;

  cmplx retval;
  cmplx exp[3];
  cmplx mass(m0);

#ifdef DEBUG
  callCount_++;
#endif

  for (int i = 0; i <= 2; ++i) {
    setlambda(-i);
    exp[i] = A0C(mass);
  }

  return Poly3T<fT>(exp[2], exp[1], exp[0], -2);
}

template <typename fT> Poly3T<fT> LTInterface::B0(fT psq, std::complex<fT> m0,
                                                  std::complex<fT> m1) {
  typedef std::complex<double> cmplx;

  cmplx retval;
  cmplx exp[3];
  double inv = psq;
  cmplx masses[2] = {m0, m1};

#ifdef DEBUG
  callCount_++;
#endif

  for (int i = 0; i <= 2; ++i) {
    setlambda(-i);
    exp[i] = B0C(inv, masses[0], masses[1]);
  }

  return Poly3T<fT>(exp[2], exp[1], exp[0], -2);
}

template <typename fT> Poly3T<fT> LTInterface::evaluate_scalar(
    const typename Invariants<fT>::Ptr & inv) {
  typedef std::complex<double> cmplx;

  typename Vector<fT>::type mominv;
  cmplx retval;
  cmplx exp[3];
  int points = inv->points();
  int numInv = (points*points - points)/2;
  double invtr[10];
  cmplx masses[5];

#ifdef DEBUG
  callCount_++;
#endif

  Interfaces::if_arguments<fT>(inv, mominv);

  for (int i = 0; i < points; ++i) {
    masses[i] = cmplx(real(inv->mass(i)), imag(inv->mass(i)));
  }

  // truncate values for specified precision, so that QL knows when
  // to switch to IR divergent evaluation

  for (int i = 0; i < numInv; ++i)
    invtr[i] = NumUtil::truncate<double>(mominv[i], musq_*true_relprec_);

  adjust_kinematics(invtr, masses, numInv, points);

  for (int i = 0; i <= 2; ++i) {
    //clearcache();
    setlambda(-i);
    exp[i] = cmplx(0,0);

    switch (inv->points()) {
      case 1: retval =  A0C(masses[0]);
              break;
      case 2: retval =  B0C(invtr[0], masses[0], masses[1]);
              break;
      case 3: retval =  C0C(invtr[0], invtr[1], invtr[2],
                           masses[0], masses[1], masses[2]);
              break;
      case 4: retval =  D0C(invtr[0], invtr[1], invtr[2], invtr[3], invtr[4],
                            invtr[5], masses[0], masses[1], masses[2],
                            masses[3]);
              break;
      case 5: retval =  E0C(invtr[0], invtr[1], invtr[2], invtr[3], invtr[4],
                            invtr[5], invtr[6], invtr[7], invtr[8], invtr[9],
                            masses[0], masses[1], masses[2],
                            masses[3], masses[4]);
              break;
    }
    exp[i] = retval;
  }

  return Poly3T<fT>(exp[2], exp[1], exp[0], -2);
}

template <typename fT> Poly3T<fT> LTInterface::dB0(fT psq,
                                                   std::complex<fT> m0,
                                                   std::complex<fT> m1) {
  typedef std::complex<double> cmplx;

  cmplx exp[3];
  cmplx masses[2] = {m0, m1};
  double inv = psq;

  for (int i = 0; i <= 2; ++i) {
    setlambda(-i);
    exp[i] = B0iC(dbb0, inv, masses[0], masses[1]);
  }

  return Poly3T<fT>(exp[2],exp[1],exp[0],-2);
}
#endif // ifdef LOOPTOOLS

#ifdef ONELOOP
template <typename fT> Poly3T<fT> OLInterface::A0(std::complex<fT> m0) {
  typedef std::complex<double> cmplx;

  cmplx res[3];

#ifdef DEBUG
  callCount_++;
#endif

// TODO: Check if OneLoop real evaluation faster after checking for imaginary part
//    if (masstri[0] == 0.)
//        OLO_A0r(&res, &masstr[0]);
//    else {
  cmplx masses(m0);
  OLO_A0c(&res, &masses);
//    }

  return Poly3T<fT>(res[2], res[1], res[0], -2);
}

template <typename fT> Poly3T<fT> OLInterface::B0(fT psq, std::complex<fT> m0,
                                                  std::complex<fT> m1) {
  typedef std::complex<double> cmplx;

  cmplx res[3];
  double invtr[1] = {psq};
  cmplx masses[2] = {m0, m1};

#ifdef DEBUG
  callCount_++;
#endif

    // TODO: Check if OneLoop real evaluation faster after checking for imaginary part

//    if (masstri[0] == 0. && masstri[1] == 0.) // Make sure OL knows there are zero widths
//        OLO_B0rr(&res, &invtr[0], &masstr[0], &masstr[1]);
//    else
  OLO_B0rc(&res, &invtr[0], &masses[0], &masses[1]);

  return Poly3T<fT>(res[2], res[1], res[0], -2);
}

template <typename fT> Poly3T<fT> OLInterface::evaluate_scalar(
    const typename Invariants<fT>::Ptr & inv) {
  typename Vector<fT>::type args;
  __complex128 resq[3];
  double invtr[6];
  __float128 invtrq[6];
  Complex128 massesq[4];
  __complex128 olmassesq[4];
  int points = inv->points();
  int numInv = (points*points - points)/2;

  Interfaces::if_arguments<fT>(inv, args);

#ifdef DEBUG
  callCount_++;
#endif

//    bool complexNeeded = false;

  for (int i = 0; i < points; ++i) {
    //massesq[i] = {real(inv->mass(i)), imag(inv->mass(i))};
    massesq[i] = inv->mass(i);
//        complexNeeded |= (masstri[i] != 0.);
  }

    // truncate values for specified precision, so that QL knows when
    // to switch to IR divergent evaluation

  for (int i = 0; i < numInv; ++i) {
    invtr[i] = NumUtil::truncate<double>(args[i], musq_*true_relprec_);
    invtrq[i] = invtr[i];
  }

  adjust_kinematics(invtrq, massesq, numInv, points);

  for (int i = 0; i < points; ++i) {
    //olmassesq[i] = {real(massesq[i]), imag(massesq[i])};
    olmassesq[i] = massesq[i].z_;
  }

 switch (points) {
    case 1: OLO_A0cq(&resq, &olmassesq[0]);
            break;
    case 2: OLO_B0rcq(&resq, &invtrq[0], &olmassesq[0], &olmassesq[1]);
            break;
    case 3: OLO_C0rcq(&resq, &invtrq[0], &invtrq[1], &invtrq[2],
                      &olmassesq[0], &olmassesq[1], &olmassesq[2]);
            break;
    case 4: OLO_D0rcq(&resq, &(invtrq[0]), &(invtrq[1]),
                      &(invtrq[2]), &(invtrq[3]),
                      &(invtrq[4]), &(invtrq[5]), 
                      &(olmassesq[0]), &(olmassesq[1]),
                      &(olmassesq[2]), &(olmassesq[3]));
            break;
  }

  return Poly3T<fT>(resq[2], resq[1], resq[0], -2);
}

template <typename fT> Poly3T<fT> OLInterface::dB0(fT psq,
                                                   std::complex<fT> m0,
                                                   std::complex<fT> m1) {
  typedef std::complex<double> cmplx;

  cmplx res[3];
  double invtr[1] = {psq};
  cmplx masses[2] = {m0, m1};

#ifdef DEBUG
  callCount_++;
#endif

  if (psq != 0. || real(masses[0]) != 0. || real(masses[1]) != 0.)
  // OneLoop doesn't handle the else case, but it's zero...
    OLO_dB0rc(&res, &invtr[0], &masses[0], &masses[1]);

  return Poly3T<fT>(res[2], res[1], res[0], -2);
}

#endif // ONELOOP

#ifdef COLLIER
template <typename fT> Poly3T<fT> CollierInterface::evaluate_tensor_coeff(
    const std::vector<int> & indices,
    const typename Invariants<fT>::Ptr & inv,
    bool & success) const {
  typedef std::complex<double> cmplx;
  int64_t callindex;
  typename Vector<fT>::type mominv;
  cmplx masses[5];
  double invtr[10];
  int points = inv->points();
  int numInv = (points*points - points)/2;
  cmplx exp[3];

  Interfaces::if_arguments<fT>(inv, mominv);

  for (int i = 0; i < inv->points(); ++i) {
    masses[i] = cmplx(real(inv->mass(i)), imag(inv->mass(i)));
  }

  // truncate values for specified precision, so that QL knows when
  // to switch to IR divergent evaluation

  for (int i = 0; i < numInv; ++i) {
    invtr[i] = NumUtil::truncate<double>(mominv[i], musq_*true_relprec_);
  }

  adjust_kinematics(invtr, masses, numInv, points);

 return Poly3T<fT>(0., 0., 0., -2);
}

template <typename fT> Poly3T<fT> CollierInterface::A0(std::complex<fT> m0) {
  typedef std::complex<double> cmplx;
    return Poly3T<fT>(0., 0., 0., -2);
}

template <typename fT> Poly3T<fT> CollierInterface::B0(fT psq, std::complex<fT> m0,
                                                  std::complex<fT> m1) {
    return Poly3T<fT>(0., 0., 0., -2);
}

template <typename fT> Poly3T<fT> CollierInterface::evaluate_scalar(
    const typename Invariants<fT>::Ptr & inv) {
    return Poly3T<fT>(0., 0., 0., -2);
}

#endif // ifdef COLLIER

template <typename fT> Poly3T<fT> PoleCheckInterface::evaluate_tensor_coeff(
    const std::vector<int> & indices,
    const typename Invariants<fT>::Ptr & inv,
    bool & success) const {
  typename Vector<fT>::type mominv;

  double sPole = 0.;
  /*double masses3[3] = {real(inv->mass(0)), real(inv->mass(1)),
                       real(inv->mass(2))};
  double masses4[4] = {real(inv->mass(0)), real(inv->mass(1)),
                       real(inv->mass(2)), real(inv->mass(3))};
  double masses5[5] = {real(inv->mass(0)), real(inv->mass(1)),
                       real(inv->mass(2)), real(inv->mass(3)),
                       real(inv->mass(4))};
*/

  double masses[5];
  for (int i = 0; i < inv->points(); ++i) {
    masses[i] = real(inv->mass(i));
  }

  Interfaces::if_arguments<fT>(inv, mominv);

  switch (inv->points()) {
    case 3:
      sPole = c_coeff(indices, mominv, masses, success);
      break;
    case 4:
      sPole = d_coeff(indices, mominv, masses, success);
      break;
    case 5:
      sPole = e_coeff(indices, mominv, masses, success);
      break;
    default:
      success = false;
  }

  return Poly3T<fT>(0., sPole, 0., -2);
}

#endif  // INTERFACES_H_
