#ifndef KINEMATICS_H_
#define KINEMATICS_H_

#include <sys/types.h>
#include <vector>
#include <map>
#include <string>
#include <utility>
#include <iostream>
#include <cstdlib>
#if __cplusplus >= 201103L
#include <memory>
using std::shared_ptr;
#include <unordered_map>
using std::unordered_map;
#else
#include <tr1/memory> // shared_ptr
using std::tr1::shared_ptr;
#include <tr1/unordered_map>
using std::tr1::unordered_map;
#endif
#include "common.h"
#include "utils.h"
#include "numerics.h"
#include "precision.h"
#include "four_vector.h"
#include "tcomplex.h"

template <typename fT> class Invariants;

/**********************************************************************
 *                                                                    *
 * KinematicsMap                                                      *
 *                                                                    *
 * Holds global state relevant for kinematics                         *
 *                                                                    *
 **********************************************************************/

// The kinematics map, having been fed the raw values from the NLOX interface,
// only stores double precision numbers for basis vectors and masses.

class KinematicsMap {
  typedef std::complex<double> cmplx;

  public:
    typedef std::vector<std::string> StringVec;
    typedef std::pair<std::string, FourVectorT<double> > StringMomPair;
    typedef std::pair<std::string, cmplx> StringMassPair;
    KinematicsMap() {}
    KinematicsMap(std::vector<StringMomPair> extmom,
                  std::vector<StringMassPair> masses);
    void new_values(std::vector<StringMomPair> extmom,
                    std::vector<StringMassPair> masses);

    template<typename fT> std::map<typename Invariants<fT>::Hash_T,
                                   typename Invariants<fT>::Ptr> & cache();

    StringVec momentum_basis() const;
    StringVec mass_basis() const;

    FourVectorT<double> basis_momentum_value(int pos) const;
    //FourVectorT<double> momentum_value(const std::vector<int> & vec) const;
    template<typename fT> FourVectorT<fT> momentum_value(const std::vector<int> & vec) const;
    cmplx mass_value(int pos) const;

    int string_to_mass(const std::string & str);
    void print_momentum(const std::vector<int> & vec, std::ostream & strm,
                        std::string prefix = "");
    void print_mass(int pos, std::ostream & strm, std::string prefix = "");
    std::vector<int> string_to_momentum(std::string str);

  private:
    StringVec momentum_names_;
    StringVec mass_names_;
    std::vector<FourVectorT<double> >  momentum_values_;
    std::vector<cmplx > mass_values_;

    // Note on Invariants caches: these have been moved from a static
    // variable inside the Invariants class here so that there is one
    // copy per KinematicsMap object (and therefore per TRed object).
    // However, the MomentumInvariants caches are left as static to
    // retain encapsulation and because it's unlikely different TRed
    // objects will be using different momenta and there is no performance
    // penalty. In fact, if the PSP is not reset, it is likely better! */

    // TODO: Implementation note on caches: in principle the different
    // precisions should share keys, and the elements should be groups
    // of Invariants of different precisions, rather than multiple caches.

    std::map<std::pair<int64_t, int64_t>,
             shared_ptr<Invariants<fT1> > > cache1_;
#if MAX_PREC_LVL >= 2
    std::map<std::pair<int64_t, int64_t>,
             shared_ptr<Invariants<fT2> > > cache2_;
#endif
#if MAX_PREC_LVL >= 3
    std::map<std::pair<int64_t, int64_t>,
             shared_ptr<Invariants<fT3> > > cache3_;
#endif

};


/**********************************************************************
 *                                                                    *
 * Invariants                                                         *
 *                                                                    *
 **********************************************************************/

template <typename fT> class Invariants {
  public:
    typedef shared_ptr<Invariants<fT> > Ptr;
    typedef shared_ptr<const Invariants<fT> > ConstPtr;
    typedef typename tcomplex<fT>::type cfT;
    typedef std::pair<int64_t, int64_t> Hash_T;

    static Invariants<fT>::Ptr get_instance(
        KinematicsMap* kmap,
        const std::vector<std::vector<int> > & momenta,
        const std::vector<int> & masses);
    static Invariants<fT>::Ptr get_instance(
        KinematicsMap* kmap,
        const std::vector<std::string> & momenta,
        const std::vector<std::string> & masses);
    ~Invariants<fT>();

    static void update_kinematics_map(KinematicsMap* kmap);
    void clear_caches();

    // creates an Invariants child by dropping a propagator
    Invariants<fT>::Ptr drop_propagator(int pos);

    static int64_t hash(const std::vector<int> & masses);
    Invariants<fT>::Hash_T get_hash() const { return hash_; }

    int points() const;

    // often needed quantities
    const typename Vector<cfT>::type & ff();
    const typename Array<cfT>::type  & x_mat();
    const typename Array<cfT>::type  & x_adj();
    cfT x_det();
    const std::vector<std::vector<typename Array<cfT>::type > > & x_gen_adj();
    const typename Array<cfT>::type & cayley_mat();

    // Accessors
    FourVectorT<fT> external_momentum(int pos) const;
    FourVectorT<fT> internal_momentum(int pos) const;
    cfT mass(int i) const;
    bool x_det_computed() const;

    // pass-throughs to MomentumInvariants
    const typename Array<fT>::type & external_invariants() const;
    const typename Array<fT>::type & internal_invariants() const;
    fT external_invariants(int i, int j) const;
    fT internal_invariants(int i, int j) const;

    const std::vector<FourVectorT<fT> > external_momenta();

    const typename Array<fT>::type & gram_mat() const;
    const typename Array<fT>::type & gram_adj() const;
    const typename Array<fT>::type & gram_adj_tr() const;
    fT gram_det() const;

    cfT cayley_minor(int r0, int c0);
    cfT cayley_minor(int r0, int r1, int c0, int c1);
    cfT cayley_minor(int r0, int r1, int r2, int c0, int c1, int c2);

    void print_external_momentum(int pos, std::ostream & strm,
                                 std::string prefix = "") const;
    void print_internal_momentum(int pos, std::ostream & strm,
                                 std::string prefix = "") const;
    void print(std::ostream & strm, std::string prefix = "") const;

  private:
    class MomentumInvariants;
      friend class MomentumInvariants;

      Invariants(KinematicsMap* kmap, MomentumInvariants & mominv,
                 const std::vector<int> & masses);

//        void sorted_invariants();

      KinematicsMap* kmap_;
      int points_;
      MomentumInvariants& mominv_;
      std::vector<int> masses_;
      bool x_det_computed_;

      std::vector<cfT> mass_values_;

      // Sorted masses and momenta by absolute value for
      // numerically stable evaluation of IR/UV poles
      // Note: These do not appear to be used anywhere.
//        std::vector<cfT> _sortedmass_values_;
//        std::vector<FourVectorT<fT> > _sorted_momenta_values;

      // caches
      typename Vector<cfT>::type ff_;
      typename Array<cfT>::type  x_mat_;
      typename Array<cfT>::type  x_adj_;
      cfT                        x_det_;
      std::vector<std::vector<typename Array<cfT>::type > > x_gen_adj_;
      typename Array<cfT>::type cayley_mat_;
      unordered_map<int64_t, cfT> cayley_minors_;
      Invariants::Hash_T hash_;

      class MomentumInvariants {
        template <typename fU> friend class Invariants;
        public:
          // Factory method
          static MomentumInvariants& get_instance(
              KinematicsMap* kmap,
              const std::vector<std::vector<int> > & momenta);
          ~MomentumInvariants();

          static void update_kinematics_map(KinematicsMap* kmap);
          void clear_caches();

          std::vector<std::vector<int> > drop_propagator(int pos);
          static int64_t hash(const std::vector<std::vector<int> > & vec);
          int64_t get_hash() { return hash_; }

          int points();

          const std::vector<FourVectorT<fT> > external_momenta();

          // often needed quantities
          const typename Array<fT>::type & external_invariants();
          const typename Array<fT>::type & internal_invariants();
          fT external_invariants(int i, int j) const;
          fT internal_invariants(int i, int j) const;

          const typename Array<fT>::type & gram_mat();
          const typename Array<fT>::type & gram_adj();
          const typename Array<fT>::type & gram_adj_tr();
          fT gram_det();

          const std::vector<std::vector<typename Array<fT>::type > >
          gram_gen_adj();

        private:
          // use factory method instead of direct ctor
          MomentumInvariants(KinematicsMap *kmap,
                             const std::vector<std::vector<int> > & momenta);

          void invariants(const std::vector<FourVectorT<fT> > & mom,
                          typename Array<fT>::type & inv);

          void calc_momenta();

          int points_;
          int64_t hash_;
          KinematicsMap* kmap_;
          std::vector<std::vector<int> > ext_momenta_;
          std::vector<std::vector<int> > int_momenta_;
          std::vector<FourVectorT<fT> > ext_momenta_values_;
          std::vector<FourVectorT<fT> > int_momenta_values_;

          // caches
          static std::map<int64_t, MomentumInvariants> cache_;

          typename Array<fT>::type ext_invs_;
          typename Array<fT>::type int_invs_;
          typename Array<fT>::type gram_mat_;
          typename Array<fT>::type gram_adj_;
          typename Array<fT>::type gram_adj_tr_;

          fT gram_det_;
          bool gram_det_computed_;
          std::vector<std::vector<typename Array<fT>::type > > gram_gen_adj_;
    };
};

// caches

template <typename fT>
typename std::map<int64_t, typename Invariants<fT>::MomentumInvariants>
Invariants<fT>::MomentumInvariants::cache_;

template <> inline std::map<Invariants<fT1>::Hash_T, Invariants<fT1>::Ptr> &
KinematicsMap::cache<fT1>() {
  return cache1_;
}
#if MAX_PREC_LVL >= 2
template <> inline std::map<Invariants<fT2>::Hash_T, Invariants<fT2>::Ptr> &
KinematicsMap::cache<fT2>() {
  return cache2_;
}
#endif
#if MAX_PREC_LVL >= 3
template <> inline std::map<Invariants<fT3>::Hash_T, Invariants<fT3>::Ptr> &
KinematicsMap::cache<fT3>() {
  return cache3_;
}
#endif

/**********************************************************************
 *                                                                    *
 * Inlined functions                                                  *
 *                                                                    *
 **********************************************************************/

inline FourVectorT<double> KinematicsMap::basis_momentum_value(int pos) const {
  return momentum_values_[pos];
}

/*inline FourVectorT<double> KinematicsMap::momentum_value(
    const std::vector<int> & vec) const {
  FourVectorT<double> res;
  for (std::vector<int>::size_type i = 0; i < vec.size(); ++i)
    res += momentum_values_[i]*((double) vec[i]);
  return res;
}*/
template <typename fT> inline FourVectorT<fT> KinematicsMap::momentum_value(const std::vector<int> & vec) const {
  FourVectorT<fT> res;
  for (std::vector<int>::size_type i = 0; i < vec.size(); ++i)
    //res += momentum_values_[i]*(fT(vec[i]));
    res += FourVectorT<fT>(momentum_values_[i])*(fT(vec[i]));
  return res;
}

inline std::complex<double> KinematicsMap::mass_value(int pos) const {
  return mass_values_[pos];
}

inline std::vector<std::string> KinematicsMap::momentum_basis() const {
  return momentum_names_;
}

inline std::vector<std::string> KinematicsMap::mass_basis() const {
  return mass_names_;
}

template <typename fT> inline typename tcomplex<fT>::type
Invariants<fT>::mass(int i) const {
  return mass_values_[i];
}

template <typename fT> inline int Invariants<fT>::points() const {
  return points_;
}

template <typename fT> inline bool Invariants<fT>::x_det_computed() const {
  return x_det_computed_;
}

template <typename fT> inline FourVectorT<fT>
Invariants<fT>::external_momentum(int pos) const {
  return mominv_.ext_momenta_values_[pos];
}

template <typename fT> inline FourVectorT<fT>
Invariants<fT>::internal_momentum(int pos) const {
  return mominv_.ext_momenta_values_[pos];
}

template <typename fT> inline const std::vector<FourVectorT<fT> >
Invariants<fT>::external_momenta() {
  return mominv_.external_momenta();
}

template <typename fT> inline const typename Array<fT>::type &
Invariants<fT>::external_invariants() const {
  return mominv_.external_invariants();
}

template <typename fT> inline fT
Invariants<fT>::external_invariants(int i, int j) const {
  return mominv_.external_invariants(i, j);
}

template <typename fT> inline const typename Array<fT>::type &
Invariants<fT>::internal_invariants() const {
  return mominv_.internal_invariants();
}

template <typename fT> inline fT
Invariants<fT>::internal_invariants(int i, int j) const {
  return mominv_.internal_invariants(i, j);
}

template <typename fT> inline const typename Array<fT>::type &
Invariants<fT>::gram_mat() const {
  return mominv_.gram_mat();
}

template <typename fT> inline const typename Array<fT>::type &
Invariants<fT>::gram_adj() const {
  return mominv_.gram_adj();
}

template <typename fT> inline const typename Array<fT>::type &
Invariants<fT>::gram_adj_tr() const {
  return mominv_.gram_adj_tr();
}

template <typename fT> inline fT Invariants<fT>::gram_det() const {
  return mominv_.gram_det();
}

template <typename fT> inline
const typename Vector<typename tcomplex<fT>::type >::type &
Invariants<fT>::ff() {
  if (ff_.size() == 0)
    for (int i = 0; i < points_-1; ++i)
      ff_.push_back(internal_invariants(i,i) + mass_values_[0]
                                             - mass_values_[i+1]);
  return ff_;
}

// TODO: Think about whether inlining a function this long is a good idea.
template <typename fT> inline
const typename Array<typename tcomplex<fT>::type >::type &
Invariants<fT>::x_mat() {
  typedef typename tcomplex<fT>::type cfT;
  if (x_mat_.size() == 0) {
    typename Vector<cfT>::type f = ff();

    for (int i = 0; i < points_; ++i) {
      std::vector<cfT> row;
      if (i == 0) {
        row.push_back(fT(2)*mass_values_[0]);
        for (int j = 0; j < points_-1; ++j)
          row.push_back(f[j]);
      } else {
        row.push_back(f[i-1]);
        for (int j = 0; j < points_-1; ++j)
          row.push_back(2*internal_invariants(i-1,j));
      }
      x_mat_.push_back(row);
    }
  }
  return x_mat_;
}

template <typename fT> inline
const typename Array<typename tcomplex<fT>::type >::type &
Invariants<fT>::x_adj() {
  if (x_adj_.size() == 0)
    x_adj_ = NumUtil::adjoint_matrix<typename tcomplex<fT>::type >(x_mat());
  return x_adj_;
}

template <typename fT> inline typename tcomplex<fT>::type
Invariants<fT>::x_det() {
  if (!x_det_computed_) {
    x_det_ = NumUtil::determinant<typename tcomplex<fT>::type >(x_mat(),
                                                                points_);
    x_det_computed_ = true;
  }
  return x_det_;
}

// TODO: typedef that extremely long return type somehow
template <typename fT> inline
const std::vector<std::vector<typename Array<typename tcomplex<fT>::type >::type > > &
Invariants<fT>::x_gen_adj() {
  if (x_gen_adj_.size() == 0) {
    NumUtil::populate_gen_adjoint<typename tcomplex<fT>::type >(x_gen_adj_,
                                                                x_mat());
  }
  return x_gen_adj_;
}

template <typename fT> inline
const typename Array<typename tcomplex<fT>::type >::type &
Invariants<fT>::cayley_mat() {
  // TODO: check whether this is correct!!!!
  if (cayley_mat_.size() == 0) {
    std::vector<FourVectorT<fT> > mom = mominv_.int_momenta_values_;
    // add the redundant momentum
    mom.push_back(FourVectorT<fT>(0, 0, 0, 0));
    for (int i = 0; i < points_ + 1; ++i) {
      typename Vector<typename tcomplex<fT>::type >::type row;
      for (int j = 0; j < points_ + 1; ++j) {
        if (i == 0) {
          if (j == 0)
            row.push_back(0);
          else
            row.push_back(1);
        } else {
          if (j == 0) {
            row.push_back(1);
          } else {
            // mpoints_^tausk == m_0
            int im = i % points_;
            int jm = j % points_;
            typename tcomplex<fT>::type val =
                - FourVectorT<fT>::scalar_product(mom[i-1]-mom[j-1],
                                                  mom[i-1]-mom[j-1])
                + mass_values_[im] + mass_values_[jm];
            row.push_back(val);
          }
        }
      }
      cayley_mat_.push_back(row);
    }
  }
  return cayley_mat_;
}

template <typename fT> inline const std::vector<FourVectorT<fT> >
Invariants<fT>::MomentumInvariants::external_momenta() {
  return ext_momenta_values_;
}

template <typename fT> inline const typename Array<fT>::type &
Invariants<fT>::MomentumInvariants::external_invariants() {
  if (ext_invs_.size() == 0)
    invariants(ext_momenta_values_, ext_invs_);
  return ext_invs_;
}

template <typename fT> inline fT
Invariants<fT>::MomentumInvariants::external_invariants(int i, int j) const {
  if (ext_invs_.size() == 0)
    return FourVectorT<fT>::scalar_product(ext_momenta_values_[i], ext_momenta_values_[j]);
  else return ext_invs_[i][j];
}

template <typename fT> inline const typename Array<fT>::type &
Invariants<fT>::MomentumInvariants::internal_invariants() {
  if (int_invs_.size() == 0)
    invariants(int_momenta_values_, int_invs_);
  return int_invs_;
}

template <typename fT> inline fT
Invariants<fT>::MomentumInvariants::internal_invariants(int i, int j) const {
  return FourVectorT<fT>::scalar_product(int_momenta_values_[i],
                                         int_momenta_values_[j]);
}

template <typename fT> inline const typename Array<fT>::type &
Invariants<fT>::MomentumInvariants::gram_mat() {
  if (gram_mat_.size() == 0) {
    for (int i = 0; i < points_-1; ++i) {
      typename Vector<fT>::type row;
      for (int j = 0; j < points_-1; ++j)
        row.push_back(2*internal_invariants(i, j));
      gram_mat_.push_back(row);
    }
  }
  return gram_mat_;
}

template <typename fT> inline const typename Array<fT>::type &
Invariants<fT>::MomentumInvariants::gram_adj() {
  if (gram_adj_.size() == 0) {
    const typename Array<fT>::type & mat = gram_mat();
    gram_adj_ = NumUtil::adjoint_matrix<fT>(mat);
  }
  return gram_adj_;
}

template <typename fT> inline const typename Array<fT>::type &
Invariants<fT>::MomentumInvariants::gram_adj_tr() {
  if (gram_adj_tr_.size() == 0) {
    const typename Array<fT>::type & pre = gram_adj();
    // transpose
    for (typename Array<fT>::size_type i = 0; i < pre.size(); ++i) {
      typename Vector<fT>::type row;
      for (typename Array<fT>::size_type j = 0; j < pre.size(); ++j)
        row.push_back(pre[j][i]);
      gram_adj_tr_.push_back(row);
    }
  }
  return gram_adj_tr_;
}

template <typename fT> inline fT
Invariants<fT>::MomentumInvariants::gram_det() {
  using namespace std; // std::pow for double, long double
  if (!gram_det_computed_) {
    gram_det_ = pow(fT(2), points_-1)
              *NumUtil::determinant<fT>(internal_invariants(), points_-1);
    gram_det_computed_ = true;
  }
  return gram_det_;
}

template <typename fT> inline int Invariants<fT>::MomentumInvariants::points() {
  return points_;
}

// non-inlined

template <typename fT> Invariants<fT>::Invariants(
    KinematicsMap* kmap, MomentumInvariants & mominv,
    const std::vector<int> & masses)
    : kmap_(kmap),
      points_(masses.size()),
      mominv_(mominv),
      masses_(masses),
      x_det_computed_(false) {
  if(points_ != mominv.points()) {
    std::cerr << "Error: wrong number of momenta or masses specified"
              << std::endl;
    exit(1);
  }
  for (std::vector<int>::size_type i = 0; i < masses_.size(); ++i) {
    std::complex<double> mvalue = kmap_->mass_value(masses_[i]);
    mass_values_.push_back(mvalue*mvalue);
  }
//    sorted_invariants();
}

template <typename fT> typename Invariants<fT>::Ptr
Invariants<fT>::get_instance(KinematicsMap* kmap,
                             const std::vector<std::vector<int> > & momenta,
                             const std::vector<int> & masses) {
  using std::make_pair;

  MomentumInvariants& mominv = MomentumInvariants::get_instance(kmap,
                                                                momenta);
  int64_t momhash  = mominv.get_hash();
  int64_t masshash = Invariants<fT>::hash(masses);

  typename std::map<Invariants<fT>::Hash_T,
                    Invariants<fT>::Ptr>::iterator found =
      kmap->cache<fT>().find(make_pair(momhash, masshash));

  if (found == kmap->cache<fT>().end()) {
    Invariants<fT>::Ptr inv(new Invariants<fT>(kmap, mominv, masses));
    found = kmap->cache<fT>().insert(make_pair(make_pair(momhash, masshash),
                                               inv)).first;
    inv->hash_ = make_pair(momhash, masshash);
  }

  return found->second;
}

template <typename fT> void Invariants<fT>::clear_caches() {
  ff_.clear();
  x_mat_.clear();
  x_adj_.clear();
  x_gen_adj_.clear();
  cayley_mat_.clear();
  cayley_minors_.clear();
  mass_values_.clear();
  x_det_computed_ = false;
}

template <typename fT> void Invariants<fT>::update_kinematics_map(
    KinematicsMap* kmap) {
  Invariants<fT>::MomentumInvariants::update_kinematics_map(kmap);
  for(typename std::map<Invariants<fT>::Hash_T,
                        Invariants<fT>::Ptr>::iterator i
      = kmap->cache<fT>().begin();
      i != kmap->cache<fT>().end(); ++i) {
    i->second->kmap_ = kmap;
    i->second->clear_caches();
//        i->second->sorted_invariants();
    // NEW: Rebuild masses, since now they can change
    for (std::vector<int>::size_type j = 0; j < i->second->masses_.size(); ++j) {
      std::complex<double> mvalue = kmap->mass_value(i->second->masses_[j]);
      i->second->mass_values_.push_back(mvalue*mvalue);
    }
  }
}

template <typename fT> typename Invariants<fT>::Ptr
Invariants<fT>::get_instance(KinematicsMap* kmap,
                             const std::vector<std::string> & momenta,
                             const std::vector<std::string> & masses) {
  std::vector<std::vector<int> > momvec;
  std::vector<int> massvec;
  for (std::vector<std::string>::const_iterator it = momenta.begin();
       it != momenta.end(); ++it) {
    std::vector<int> mom = kmap->string_to_momentum(*it);
    momvec.push_back(mom);
  }
  for (std::vector<std::string>::const_iterator it = masses.begin();
       it != masses.end(); ++it) {
    int mass = kmap->string_to_mass(*it);
    massvec.push_back(mass);
  }
  return Invariants<fT>::get_instance(kmap, momvec, massvec);
}

template <typename fT> int64_t Invariants<fT>::hash(
    const std::vector<int> & masses) {
  int64_t hash = 0;
  for (std::vector<int>::const_iterator it = masses.begin();
       it != masses.end(); ++it)
    hash |= (*it + 1) << ((it - masses.begin())*4);
  return hash;
}

template <typename fT> Invariants<fT>::~Invariants<fT>() {}

template <typename fT> typename Invariants<fT>::Ptr
Invariants<fT>::drop_propagator(int pos) {
  std::vector<std::vector<int> > newmoms = mominv_.drop_propagator(pos);
  std::vector<int> newmasses;
  for (std::vector<int>::size_type i = 0; i < masses_.size(); ++i) {
    if (pos != 0) {
      if (i != (unsigned int) pos)
        newmasses.push_back(masses_[i]);
    } else if (i < (unsigned int) (points_-1)) {
        newmasses.push_back(masses_[(i+1)]);
    }
  }

  return Invariants<fT>::get_instance(kmap_, newmoms, newmasses);
}
/*
template <typename fT>
void Invariants<fT>::sorted_invariants() {
#ifdef DEBUG
    std::cout << "SORTING INVARIANTS" << std::endl;
#endif
    _sortedmass_values_ = mass_values_;
    _sorted_momenta_values = external_momenta(); 
#ifdef DEBUG
    std::cout << "MASSES: size " << mass_values_.size() << std::endl;
    for(int k = 0; k < mass_values_.size(); ++k)
        std::cout << mass_values_[k] << std::endl;
    std::cout << "MOMENTA: size " << _sorted_momenta_values.size()
              << std::endl;
    for(int k = 0; k < _sorted_momenta_values.size(); ++k)
        _sorted_momenta_values[k].print(std::cout);
#endif
    std::vector<fT> values(_sorted_momenta_values.size());
    std::transform(_sorted_momenta_values.begin(), 
                   _sorted_momenta_values.end(), 
                   values.begin(),
                   std::mem_fun_ref(&FourVectorT<fT>::abs));

    std::vector<fT> realmasses;
    for (int k = 0; k < mass_values_.size(); ++k)
        realmasses.push_back(real(mass_values_[k]));

    fT maxmominvs = Utils::max_abs_element(values);
    fT maxmassinvs = Utils::max_abs_element(realmasses);
    fT maxvalue = std::max(maxmominvs, maxmassinvs);

    if(Utils::n_elements_equal_to<fT>(realmasses, 0.,
                                      maxvalue*1e-15) == points_) {
        std::vector<int> order = 
            Utils::order_by_zero_groups(values, maxvalue*1e-15);
        if(order.size() > 0) 
            Utils::reorder(_sorted_momenta_values, order);
    } else {
        std::vector<int> order = 
            Utils::order_by_zero_groups(realmasses, maxvalue*1e-15);
        if(order.size() > 0) {
            Utils::reorder(_sortedmass_values_, order);
#ifdef DEBUG
            for(int k = 0; k < _sorted_momenta_values.size(); ++k) {
                std::cout << mominv_.ext_momenta_[k] << std::endl;
                _sorted_momenta_values[k].print(std::cout);
            }
            std::cout << "sizes: " << _sorted_momenta_values.size() << " "
                      << order.size() << std::endl;
#endif
            Utils::reorder(_sorted_momenta_values, order);
        }
    }
}
*/
template <typename fT> void Invariants<fT>::print_external_momentum(
    int pos, std::ostream & strm, std::string prefix) const {
  if(mominv_.ext_momenta_.size() == 0)
    return;
  kmap_->print_momentum(mominv_.ext_momenta_[pos], strm, prefix);
}

template <typename fT> void Invariants<fT>::print_internal_momentum(
    int pos, std::ostream & strm, std::string prefix) const {
  kmap_->print_momentum(mominv_.int_momenta_[pos], strm, prefix);
}

template <typename fT> void Invariants<fT>::print(std::ostream & strm,
                                                  std::string prefix) const {
  strm << prefix << "external momenta" << std::endl;
  for (int i = 0; i < points_; ++i) {
    print_external_momentum(i, strm, prefix);
    strm << prefix << " -- ";
    kmap_->print_mass(masses_[i], strm, prefix);
    strm << std::endl << "   ";
    if(mominv_.ext_momenta_.size() != 0)
      mominv_.ext_momenta_values_[i].print(strm, prefix);
  }
  strm << std::endl;
  strm << prefix << "internal momenta" << std::endl;
  for (int i = 0; i < points_-1; ++i) {
    print_internal_momentum(i, strm, prefix);
    strm << std::endl << prefix << "   ";
    mominv_.int_momenta_values_[i].print(strm, prefix);
  }
  strm << std::endl;
}


/**********************************************************************
 *                                                                    *
 * Invariants::MomentumInvariants                                     *
 *                                                                    *
 **********************************************************************/

template <typename fT>
Invariants<fT>::MomentumInvariants::MomentumInvariants(
    KinematicsMap* kmap,
    const std::vector<std::vector<int> > & momenta)
    : points_(momenta.size()+1), kmap_(kmap), ext_momenta_(momenta),
              gram_det_computed_(false) {
  calc_momenta();
  hash_ = MomentumInvariants::hash(momenta);
}

template <typename fT>
Invariants<fT>::MomentumInvariants::~MomentumInvariants() {}

template <typename fT> typename Invariants<fT>::MomentumInvariants &
Invariants<fT>::MomentumInvariants::get_instance(
    KinematicsMap* kmap, const std::vector<std::vector<int> > & momenta) {
  int64_t hash = MomentumInvariants::hash(momenta);
  typename std::map<int64_t, MomentumInvariants>::iterator found =
    MomentumInvariants::cache_.find(hash);

  if (found == MomentumInvariants::cache_.end()) {
    found = MomentumInvariants::cache_.insert(std::make_pair(
        hash, MomentumInvariants(kmap, momenta))).first;
  }
// TODO: Check if this is needed, turn off by default if not
#ifndef NDEBUG
  // check if there is a hash collision by comparing found momenta
  // and new momenta
  if (found != MomentumInvariants::cache_.end()) {
    if(found->second.ext_momenta_.size() != 0) {
      std::vector<std::vector<int> > this_ext_indpendent_momenta(
          found->second.ext_momenta_.begin(),
          found->second.ext_momenta_.end()-1);
      if (this_ext_indpendent_momenta != momenta) {
        std::cerr << "FATAL Error: hash collision:" << std::endl
                  << "Cache returned wrong MomentumInvariants"
                  << std::endl
                  << "found: " << this_ext_indpendent_momenta
                  << "reqs:  " << momenta;
        exit(1);
      }
    }
  }
#endif
  return found->second;
}

template <typename fT> void
Invariants<fT>::MomentumInvariants::update_kinematics_map(KinematicsMap* kmap) {
  for(typename std::map<int64_t, MomentumInvariants>::iterator i = cache_.begin();
      i != cache_.end(); ++i) {
    i->second.kmap_ = kmap;
    i->second.clear_caches();
    i->second.calc_momenta();
  }
}

template <typename fT> void Invariants<fT>::MomentumInvariants::clear_caches() {
  ext_momenta_values_.clear();
  int_momenta_values_.clear();
  ext_invs_.clear();
  int_invs_.clear();
  gram_mat_.clear();
  gram_adj_.clear();
  gram_adj_tr_.clear();
  gram_gen_adj_.clear();
  gram_det_computed_ = false;
}

template <typename fT> void Invariants<fT>::MomentumInvariants::calc_momenta() {
  if(ext_momenta_.size() == 0)
    return;

  // cache FourVectorT values for external momenta and compute the
  // dependent external momentum
  std::vector<int> lastmom(ext_momenta_[0].size(),0);
  for (std::vector<std::vector<int> >::iterator i = ext_momenta_.begin();
       i != ext_momenta_.end(); ++i) {
//    ext_momenta_values_.push_back(kmap_->momentum_value(*i));
    ext_momenta_values_.push_back(kmap_->momentum_value<fT>(*i));
    lastmom -= *i;
  }
  if(ext_momenta_values_.size() != points_)
    //ext_momenta_values_.push_back(kmap_->momentum_value(lastmom));
    ext_momenta_values_.push_back(kmap_->momentum_value<fT>(lastmom));
  if(ext_momenta_.size() != points_)
    ext_momenta_.push_back(lastmom);

  for (int i = 0; i < points_-1; ++i) {
    std::vector<int> newmom(ext_momenta_[0].size(), 0);
    for (int j = 0; j < i+1; ++j) {
      newmom += ext_momenta_[j];
    }
    if(int_momenta_.size() != points_ - 1)
      int_momenta_.push_back(newmom);
    //int_momenta_values_.push_back(kmap_->momentum_value(newmom));
    int_momenta_values_.push_back(kmap_->momentum_value<fT>(newmom));
  }

#ifdef DEBUG
  std::cout << "momenta and values size check!" << std::endl;
  std::cout << ext_momenta_.size() << ", " << ext_momenta_values_.size()
            << ", " << points_ << std::endl;
#endif
}

template <typename fT> void Invariants<fT>::MomentumInvariants::invariants(
    const std::vector<FourVectorT<fT> > & mom,
    typename Array<fT>::type & inv) {
  for (typename std::vector<FourVectorT<fT> >::size_type i = 0;
       i < mom.size(); ++i) {
    typename Vector<fT>::type row;
    for (typename std::vector<FourVectorT<fT> >::size_type j = 0;
         j < mom.size(); ++j)
      row.push_back(FourVectorT<fT>::scalar_product(mom[i], mom[j]));
    inv.push_back(row);
  }
}

template <typename fT> std::vector<std::vector<int> >
Invariants<fT>::MomentumInvariants::drop_propagator(int pos) {
  std::vector<std::vector<int> > newmoms;
  if (points_ == 2) {
    //assert(ext_momenta_[0].size() != 0);
    return newmoms;
  }
  if (pos > 0) {
    for (int i = 0; i < points_-1; ++i) {
      if (i > pos-1) {
        if (i == pos)
          newmoms.push_back(ext_momenta_[i] + ext_momenta_[pos-1]);
        else
          newmoms.push_back(ext_momenta_[i]);
      } else if (i < pos-1) {
        newmoms.push_back(ext_momenta_[i]);
      }
    }
  } else if (pos == 0) {
    for (int i = 1; i < points_-1; ++i) {
      newmoms.push_back(ext_momenta_[i]);
    }
  }
  return newmoms;
}

template <typename fT> int64_t Invariants<fT>::MomentumInvariants::hash(
    const std::vector<std::vector<int> > & vec) {
  int64_t hash = 0;
  if(vec.size() == 0)
    return 0;

  int count = 0;
  int ss = 0;
  for (std::vector<std::vector<int> >::const_iterator i = vec.begin();
       i != vec.end(); ++i) {
    // momenta
    for (std::vector<int>::size_type j = 0; j < i->size(); ++j) {
      if ((*i)[j] == 0) {
        count++;
        continue;
      }
      else if ((*i)[j] == 1) {
        int64_t oldhash = hash;
        int64_t one = 1;
        hash |= (static_cast<int64_t>(one) << (2*(count)));
        if(hash == oldhash) {
          std::cerr << "HASH PROBLEM" << std::endl;
        }
      }
      else if ((*i)[j] == -1) {
        int64_t oldhash = hash;
        int64_t two = 2;
        hash |= (static_cast<int64_t>(two) << (2*(count)));
        if(hash == oldhash) {
          std::cerr << "HASH PROBLEM" << std::endl;
        }
      }
      count++;
    }
  }
  hash |= static_cast<int64_t>(0xf) << (2*count+4);
  return hash;
}

template <typename fT> typename tcomplex<fT>::type
Invariants<fT>::cayley_minor(int r0, int c0) {
  typedef typename tcomplex<fT>::type cfT;

  int64_t hash = (r0 + 1)*1000000 + (c0 + 1)*100000;
  typename unordered_map<int64_t, cfT>::iterator found =
      cayley_minors_.find(hash);
  if(found != cayley_minors_.end())
    return found->second;
  cfT res = NumUtil::general_minor<cfT>(cayley_mat(),
                                        std::vector<int>(1,r0),
                                        std::vector<int>(1,c0));
  cayley_minors_[hash] = res;
  return res;
}

template <typename fT> typename tcomplex<fT>::type
Invariants<fT>::cayley_minor(int r0, int r1, int c0, int c1) {
  typedef typename tcomplex<fT>::type cfT;

  int64_t hash = (r0 + 1)*1000000 + (c0 + 1)*100000
               + (r1+1)*10000 + (c1+1)*1000;
  typename unordered_map<int64_t, cfT>::iterator found =
    cayley_minors_.find(hash);
  if(found != cayley_minors_.end())
    return found->second;

  int rows[] = {r0, r1};
  int cols[] = {c0, c1};
  cfT res = NumUtil::general_minor<cfT>(cayley_mat(),
                                        std::vector<int>(rows, rows+2),
                                        std::vector<int>(cols, cols+2));
  cayley_minors_[hash] = res;
  return res;
}

template <typename fT> typename tcomplex<fT>::type
Invariants<fT>::cayley_minor(int r0, int r1, int r2,
                             int c0, int c1, int c2) {
  int64_t hash = (r0 + 1)*1000000 + (c0 + 1)*100000
               + (r1+1)*10000 + (c1+1)*1000
               + (r2 + 1)*100 + (c2+1)*10;
  typename unordered_map<int64_t, cfT>::iterator found
      = cayley_minors_.find(hash);
  if(found != cayley_minors_.end())
    return found->second;
  int rows[] = {r0, r1, r2};
  int cols[] = {c0, c1, c2};
  cfT res = NumUtil::general_minor<cfT>(cayley_mat(),
                                        std::vector<int>(rows, rows+3),
                                        std::vector<int>(cols, cols+3));
  cayley_minors_[hash] = res;
  return res;
}

#endif  // KINEMATICS_H_
