#ifndef NUMERICS_H_
#define NUMERICS_H_

#include <algorithm>
#include <cstdarg>
#include <complex>
#include <iostream>
#include <vector>
#include <iomanip>
#include <math.h> // std::fabs
#include "common.h"
#include "poly3.h"
#include "precision.h"

/**********************************************************************
 *                                                                    *
 * NumUtil                                                            *
 *                                                                    *
 **********************************************************************/

/**
 * Utility functions for matrix handling and simplified handling of
 * complex structures
 */

namespace NumUtil {
  template <typename fT> int compare(fT d1, fT d2);

  template <typename fT> fT binomial(int n, int k);
  int factorial(int n);
  // computes powers of -1 efficiently and exactly ((-1)^n)
  int signpow(int n);

  template <typename fT> fT truncate(fT value, fT absprec);
  template <typename fT> fT sorted_sum(typename Vector<fT>::type & vec);
  template <typename fT> tcomplex<fT> sum_cmplx_array(tcomplex<fT> *a,
                                                      int size);
  template <typename fT> Poly3T<fT> sum_Poly3_array(Poly3T<fT> *a, int size);
  template <typename fT> fT sum_real_array(fT *a, int size);

  int signum(int a);
  template <typename fT> typename Array<fT>::type minor_matrix(
      const typename Array<fT>::type & mat, int row, int col);
  template <typename fT> typename Array<fT>::type minor_minor_matrix(
      const typename Array<fT>::type & mat,
      int row1, int row2, int col1, int col2);

  // gives the signature of the permutation needed to place the
  // elements of list in canonical order and sorts the list
  int signature_sort(std::vector<int> & list);
  template <typename fT> fT general_minor(const typename Array<fT>::type & mat,
                                          std::vector<int> rows,
                                          std::vector<int> cols);

  template <typename fT> typename Array<fT>::type adjoint_matrix(
      const typename Array<fT>::type & mat);

  template <typename fT>
  fT adjoint_element(const typename Array<fT>::type & mat, int i, int j);
  template <typename fT> fT general_adjoint_element(
      const typename Array<fT>::type & mat, int i, int j, int k, int l);
  template <typename fT> fT determinant(const typename Array<fT>::type & mat,
                                        int dim);
  // Computes the square of the volume of matrix if the rows were perpendicular
  // Useful for determining whether a determinant is small
  template <typename fT> fT ortho_vol_sq(const typename Array<fT>::type & mat,
                                         int dim);

  void populate_vector(std::vector<int> & vec, int size, ...);
  template <typename fT> void populate_gen_adjoint(
      std::vector<std::vector<typename Array<fT>::type> > & adj,
      const typename Array<fT>::type & mat);

  template <typename fT> bool equal_float(fT v1, fT v2, fT relprec, fT absprec);
  template <typename fT> bool is_nan(fT val);
  template <typename fT> bool zero_float(fT v1, fT v2, fT relprec);
}

// Inline and template functions for NumUtil

// Note for binomial: numerator/denominator computation split to minimize
// number of divisions for speed. For large integers this will result in
// overflow! Restore division in loop or change to floating type if problems.
template <typename fT> fT NumUtil::binomial(int n, int k) {
  if (n == 0 && k != 0)
    return 0;
  if (k == 0 || k == n)
    return 1;
  int kk = std::min(k, n-k);
  int res = n-kk+1;
  int den = 1;
  for (int i = 2; i <= kk; ++i) {
    res *= n-kk+i;
    den *= i;
  }
  return res/fT(den);
}

template <typename fT> fT NumUtil::truncate(fT value, fT absprec) {
//  using namespace std;
  if (fabs(value) < absprec) {
    return 0.;
  } else {
    return value;
  }
}

template <typename fT> bool NumUtil::equal_float(fT v1, fT v2,
                                                 fT relprec, fT absprec) {
//  using namespace std;
  if (fabs(v1-v2) < absprec)
    return true;
  float relError;
  if (fabs(v2) > fabs(v1))
    relError = fabs((v1 - v2) / v2);
  else
    relError = fabs((v1 - v2) / v1);
  if (relError <= relprec)
    return true;
  return false;
}

template <typename fT> fT NumUtil::sorted_sum(typename Vector<fT>::type & vec) {
  typename Vector<fT>::type sortvec = vec;
  sort(sortvec.begin(), sortvec.end(), NumUtil::compare);

  fT res = sortvec[0];
  for (typename Vector<fT>::size_type i = 1; i < sortvec.size(); ++i)
    res += sortvec[i];

  return res;
}

template <typename fT> fT NumUtil::sum_real_array(fT *a, int size) {
  fT res = 0.;
  fT s[size];
  for (int i = 0; i < size; ++i)
    s[i] = a[i];

  std::sort(s, s+size, NumUtil::compare);

  for (int i = 0; i < size; ++i)
    res += s[i];

  return res;
}

template <typename fT> typename tcomplex<fT>::type NumUtil::sum_cmplx_array(
    tcomplex<fT> *a, int size) {
  typedef typename tcomplex<fT>::type cfT;
  cfT res = cfT(0, 0);
  fT sr[size], si[size];
  for (int i = 0; i < size; ++i) {
    sr[i] = real(a[i]);
    si[i] = imag(a[i]);
  }
  std::sort(sr, sr+size, NumUtil::compare);
  std::sort(si, si+size, NumUtil::compare);

  for (int i = 0; i < size; ++i)
    res += cmplx(sr[i], si[i]);

  return res;
}

template <typename fT> int NumUtil::compare(fT d1, fT d2) {
  return fabs(d1) > fabs(d2);
}

template <typename fT> typename Array<fT>::type NumUtil::minor_matrix(
    const typename Array<fT>::type & mat, int row, int col) {
  int dim = mat.size();

  std::vector<std::vector<fT> > submat(dim-1);
  for (int j = 0; j < row; ++j) {
    std::vector<fT> rowvec(dim-1);
    for (int k = 0; k < col; ++k)
      rowvec[k] = mat[j][k];
    for (int k = col; k < dim-1; ++k)
      rowvec[k] = mat[j][k+1];
    submat[j] = rowvec;
  }
  for (int j = row; j < dim-1; ++j) {
    std::vector<fT> rowvec(dim-1);
    for (int k = 0; k < col; ++k)
      rowvec[k] = mat[j+1][k];
    for (int k = col; k < dim-1; ++k)
      rowvec[k] = mat[j+1][k+1];
    submat[j] = rowvec;
  }

  return submat;
}

// If we only need double minor, compute directly with this function for speed
// Assumed row1 < row2, col1 < col2, row2 and col2 have been decremented 1
template <typename fT> typename Array<fT>::type NumUtil::minor_minor_matrix(
    const typename Array<fT>::type & mat,
    int row1, int row2, int col1, int col2) {
  int dim = mat.size()-2;

  std::vector<std::vector<fT> > submat(dim);
  for (int j = 0; j < row1; ++j) {
    std::vector<fT> rowvec(dim);
    for (int k = 0; k < col1; ++k)
      rowvec[k] = mat[j][k];
    for (int k = col1; k < col2; ++k)
      rowvec[k] = mat[j][k+1];
    for (int k = col2; k < dim; ++k)
      rowvec[k] = mat[j][k+2];
    submat[j] = rowvec;
  }
  for (int j = row1; j < row2; ++j) {
    std::vector<fT> rowvec(dim);
    for (int k = 0; k < col1; ++k)
      rowvec[k] = mat[j+1][k];
    for (int k = col1; k < col2; ++k)
      rowvec[k] = mat[j+1][k+1];
    for (int k = col2; k < dim; ++k)
      rowvec[k] = mat[j+1][k+2];
    submat[j] = rowvec;
  }
  for (int j = row2; j < dim; ++j) {
    std::vector<fT> rowvec(dim);
    for (int k = 0; k < col1; ++k)
      rowvec[k] = mat[j+2][k];
    for (int k = col1; k < col2; ++k)
      rowvec[k] = mat[j+2][k+1];
    for (int k = col2; k < dim; ++k)
      rowvec[k] = mat[j+2][k+2];
    submat[j] = rowvec;
  }

  return submat;
}

template <typename fT> fT NumUtil::determinant(
    const typename Array<fT>::type & mat, int dim) {
  using namespace std;
  if (dim == 1)
    return mat[0][0];
  if (dim == 2)
    return mat[0][0]*mat[1][1] - mat[0][1]*mat[1][0];

  fT res = 0;

  // Allocate temporary array with pointers for efficient swaps
  fT **u = new fT*[dim];
  res = 1.;
  for (int i = 0; i < dim; ++i) {
    u[i] = new fT[dim];
    for (int j = 0; j < dim; ++j) {
      u[i][j] = mat[i][j];
    }
  }

  // Change: use Gaussian elimination to compute determinant more efficiently

  for (int i = 0; i < dim; ++i) {
    double max = abs(u[i][i]);
    int maxInd = i;
    for (int j = i+1; j < dim; ++j) {
      double val = abs(u[j][i]);
      if (val > max) {
        max = val;
        maxInd = j;
      }
    }
    if (max == 0.) {
      res = 0.;
      break;
    }
    else {
      if (maxInd != i) { // swap rows
        fT *row = u[i];
        u[i] = u[maxInd];
        u[maxInd] = row;
        res *= -1;
      }
      res *= u[i][i];
      for (int j = i+1; j < dim; ++j) {
        fT factor = u[j][i]/u[i][i];
        // don't bother computing the stuff that we know is zero and never used, start with k=i+1
        for (int k = i+1; k < dim; ++k) {
          u[j][k] -= u[i][k]*factor;
        }
      }
    }
  }

  for (int i = 0; i < dim; ++i)
    delete[] u[i];
  delete[] u;

  return res;
}

template <typename fT> fT NumUtil::ortho_vol_sq(
    const typename Array<fT>::type & mat, int dim) {
  fT vol_sq = fT(1);
  for (int i = 0; i < dim; ++i) {
    fT norm_sq = fT(0);
    for (int j = 0; j < dim; ++j) norm_sq += mat[i][j]*mat[i][j];
    vol_sq *= norm_sq;
  }
  return vol_sq;
}

template <typename fT> fT NumUtil::adjoint_element(
    const typename Array<fT>::type & mat, int i, int j) {
  int dim = mat.size();
  typename Array<fT>::type minor = NumUtil::minor_matrix<fT>(mat, i, j);
  return NumUtil::determinant<fT>(minor, dim-1)*fT(signpow(i+j));
}

template <typename fT> typename Array<fT>::type NumUtil::adjoint_matrix(
    const typename Array<fT>::type & mat) {
  typename Array<fT>::type adj;
  int dim = mat.size();

  if (dim == 1) {
    adj.push_back(typename Vector<fT>::type(1, 1));
    return adj;
  }

  for (int i = 0; i < dim; ++i) {
    std::vector<fT> rowvec;
    for (int k = 0; k < dim; ++k)
      rowvec.push_back(NumUtil::adjoint_element<fT>(mat, i, k));
    adj.push_back(rowvec);
  }
  return adj;
}

template <typename fT> fT NumUtil::general_adjoint_element(
    const typename Array<fT>::type & mat, int i, int k, int j, int l) {
  if (mat.size() == 1 || i == k || j == l)
    return 0.;

//    if ( i == k || j == l)
//        return 0.;

  int row1, row2, col1, col2;

  if (k > i) {
    row1 = i;
    row2 = k-1;
  } else {
    row1 = k;
    row2 = i-1;
  }

  if (j > l) {
    col1 = l;
    col2 = j-1;
  } else {
    col1 = j;
    col2 = l-1;
  }

  typename Array<fT>::type genminor = NumUtil::minor_minor_matrix<fT>(mat,
                                                                      row1,
                                                                      row2,
                                                                      col1,
                                                                      col2);

  fT res = fT(signpow(i+k+j+l)*NumUtil::signum(i-k)*NumUtil::signum(l-j))
         *NumUtil::determinant<fT>(genminor, genminor.size());
  return res;
}

template <typename fT> fT NumUtil::general_minor(
    const typename Array<fT>::type & mat, std::vector<int> rows,
    std::vector<int> cols) {
  typename Array<fT>::size_type matDim = mat.size();
  if (matDim != 0 && matDim != mat[0].size()) {
    std::cerr << "error: NumUtil::general_minor_matrix()" << std::endl
              << "only square matrices allowed" << std::endl;
    exit(1);
  }

  int nRowsDropped = rows.size();
  int nColsDropped = cols.size();

// TODO: Check if speed improves by dropping this check, then put in DEBUG
  if (nRowsDropped != nColsDropped) {
    std::cerr << "error: NumUtil::general_minor_matrix()" << std::endl
              << "number of discarded rows and columns is not equal"
              << std::endl;
    exit(1);
  }

  int rowsig = signature_sort(rows);
  int colsig = signature_sort(cols);

  if (unique(rows.begin(), rows.end()) != rows.end())
    return 0.;
  if (unique(cols.begin(), cols.end()) != cols.end())
    return 0.;

  // full signature
  int sigsum = 0;
  for (int i = 0; i < nRowsDropped; ++i) {
    sigsum += rows[i] + cols[i];
  }
  int sig = signpow(sigsum)*colsig*rowsig;

  // build up the minor matrix
  std::vector<int> takerows;
  std::vector<int> takecols;
  for (typename Array<fT>::size_type i = 0; i < matDim; ++i) {
    if (find(rows.begin(), rows.end(), i) == rows.end())
      takerows.push_back(i);
    if (find(cols.begin(), cols.end(), i) == cols.end())
      takecols.push_back(i);
  }

  typename Array<fT>::type minorMatrix;
  for (typename Array<fT>::size_type i = 0; i < matDim-nRowsDropped; ++i) {
    typename Vector<fT>::type matrow;
    for (typename Array<fT>::size_type j = 0; j < matDim-nColsDropped; ++j)
      matrow.push_back(mat[takerows[i]][takecols[j]]);
    minorMatrix.push_back(matrow);
  }
  return fT(sig)*determinant<fT>(minorMatrix, minorMatrix.size());
}

template <typename fT> void NumUtil::populate_gen_adjoint(
    std::vector<std::vector<typename Array<fT>::type> > & adj,
    const typename Array<fT>::type & mat) {
  int dim = mat.size();

// Change: only bother computing antisymmetric indices once (and set diagonal to zero)
  for (int i = 0; i < dim; ++i) {
    std::vector<typename Array<fT>::type> vec1(dim);
    for (int j = 0; j < i; ++j) {
      typename Array<fT>::type vec2(dim);
      for (int k = 0; k < dim; ++k) {
        typename Vector<fT>::type vec3(dim);
        for (int l = 0; l < dim; ++l)
          vec3[l] = -adj[j][i][k][l];
        vec2[k] = vec3;
      }
      vec1[j] = vec2;
    }
    typename Array<fT>::type vec2(dim);
    for (int k = 0; k < dim; ++k) {
      typename Vector<fT>::type vec3(dim); // auto initialized as zero
      vec2[k] = vec3;
    }
    vec1[i] = vec2;
    for (int j = i+1; j < dim; ++j) {
      typename Array<fT>::type vec2(dim);
      for (int k = 0; k < dim; ++k) {
        typename Vector<fT>::type vec3(dim);
        for (int l = 0; l < k; ++l)
          vec3[l] = -vec2[l][k];
        vec3[k] = 0.;
        for (int l = k+1; l < dim; ++l)
          vec3[l] = NumUtil::general_adjoint_element<fT>(mat,i,j,k,l);
        vec2[k] = vec3;
      }

      vec1[j] = vec2;
    }

    adj.push_back(vec1);
  }
}

template <typename fT> bool NumUtil::zero_float(fT v1, fT v2, fT relprec) {
  using namespace std;

  fT relError = 1.;
  if (v2 != 0.)
    relError = fabs(v1/v2);
  if (relError <= relprec)
    return true;
  return false;
}


// Helper for vector<int> arithmetics
inline std::vector<int> & operator+=(std::vector<int> & lhs,
                                     const std::vector<int> & rhs) {
  for (std::vector<int>::size_type i = 0; i < lhs.size(); ++i)
    lhs[i] += rhs[i];
  return lhs;
}

inline std::vector<int> & operator-=(std::vector<int> & lhs,
                                     const std::vector<int> & rhs) {
  for (std::vector<int>::size_type i = 0; i < lhs.size(); ++i)
    lhs[i] -= rhs[i];
  return lhs;
}

inline std::vector<int> operator+(std::vector<int> lhs,
                                  const std::vector<int> rhs) {
  std::vector<int> vec = lhs;
  vec += rhs;
  return vec;
}


inline std::vector<int> operator-(std::vector<int> lhs,
                                  const std::vector<int> rhs) {
  std::vector<int> vec = lhs;
  vec -= rhs;
  return vec;
}

inline std::vector<int> & operator*=(std::vector<int> & lhs, const int rhs) {
  for (std::vector<int>::size_type i = 0; i < lhs.size(); ++i)
    lhs[i] = rhs*lhs[i];
  return lhs;
}

inline std::vector<int> operator*(const std::vector<int> lhs, const int rhs) {
  std::vector<int> vec =  lhs;
  vec *= rhs;
  return vec;
}

template <typename fT> Poly3T<fT> NumUtil::sum_Poly3_array(Poly3T<fT> *a,
                                                           int size) {
  typedef tcomplex<fT> cmplx;
  cmplx epores[3];
  for (int j = -2; j < 1; ++j) {
    epores[j+2] = cmplx(0, 0);
    fT sr[size], si[size];
    for (int i = 0; i < size; ++i) {
      sr[i] = real(a[i].order(j));
      si[i] = imag(a[i].order(j));
    }
    std::sort(sr, sr+size, NumUtil::compare);
    std::sort(si, si+size, NumUtil::compare);

    for (int i = 0; i < size; ++i)
      epores[j+2] += cmplx(sr[i], si[i]);
  }

  return Poly3T<fT>(epores[0], epores[1], epores[2], -2);
}

#endif  // NUMERICS_H_
