#ifndef TCOMPLEX_H_
#define TCOMPLEX_H_

#include <complex>
#include <iostream> // defining Complex128 stream behavior
#include "precision.h"

// Type definitions for multiple complex precision. Quadmath doesn't extend
// std::complex, so we need additional code for that.

#if MAX_PREC_LVL >= 3
#include <quadmath.h>

class Complex128 {
  public:
    Complex128(__complex128 zin = 0.Q) : z_(zin) {}
    Complex128(__float128 xin, __float128 yin = 0.Q) {
      __real__ z_ = xin;
      __imag__ z_ = yin;
    }
    template <typename fT> Complex128(std::complex<fT> zin) {
      __real__ z_ = zin.real();
      __imag__ z_ = zin.imag();
    }

    Complex128 & operator+=(const Complex128 &other) {
      z_ += other.z_;
      return *this;
    }
    Complex128 & operator-=(const Complex128 &other) {
      z_ -= other.z_;
      return *this;
    }
    Complex128 & operator*=(const Complex128 &other) {
      z_ *= other.z_;
      return *this;
    }
    Complex128 & operator/=(const Complex128 &other) {
      z_ /= other.z_;
      return *this;
    }

    __float128 real() const { return crealq(z_); }
    __float128 imag() const { return cimagq(z_); }

    __complex128 z_;
};

inline Complex128 operator-(const Complex128 &z1) {
  return Complex128(-z1.z_);
}

inline bool operator==(const Complex128 &z1, const Complex128 &z2) {
  return z1.z_ == z2.z_;
}

inline bool operator!=(const Complex128 &z1, const Complex128 &z2) {
  return z1.z_ != z2.z_;
}

inline Complex128 operator+(const Complex128 &z1, const Complex128 &z2) {
  return Complex128(z1.z_ + z2.z_);
}

inline Complex128 operator-(const Complex128 &z1, const Complex128 &z2) {
  return Complex128(z1.z_ - z2.z_);
}

inline Complex128 operator*(const Complex128 &z1, const Complex128 &z2) {
  return Complex128(z1.z_ * z2.z_);
}

inline Complex128 operator/(const Complex128 &z1, const Complex128 &z2) {
  return Complex128(z1.z_ / z2.z_);
}

inline std::ostream & operator << (std::ostream &out, const Complex128 &z) {
  out << static_cast<long double>(z.real()) << "+i"
      << static_cast<long double>(z.imag()) << std::endl;
  return out;
}

inline __float128 real(const Complex128 &z) { return z.real(); }
inline __float128 imag(const Complex128 &z) { return z.imag(); }
inline __float128 abs(const Complex128 &z) { return cabsq(z.z_); }
inline Complex128 log(const Complex128 &z) { return clogq(z.z_); }
inline Complex128 pow(const Complex128 &base, __float128 exp) {
  return cpowq(base.z_, exp);
}
inline Complex128 sqrt(const Complex128 &z) { return csqrtq(z.z_); }
#endif // MAX_PREC_LVL >= 3

template <typename fT> struct tcomplex {
  typedef typename std::complex<fT> type;
};

#if MAX_PREC_LVL >= 3
template <> struct tcomplex<__float128> {
  typedef Complex128 type;
};

#endif // MAX_PREC_LVL >= 3

#endif  // TCOMPLEX_H_
