#ifndef TRED_H_
#define TRED_H_

#include <iostream>
#include <vector>
#include <string>
#include <fstream>
#include <sstream>
#include <iterator>
#include <map>
#include <cstdarg>

#include "evaluation.h"
#include "ab_node.h" // Allows setting of zerocutoff_
#include "methods.h" // Definition of reduction method enum

class TRed {
  typedef std::complex<double> cmplx;

  public:
    typedef std::pair<std::string, FourVectorT<double> > StringMomPair;
    typedef std::pair<std::string, cmplx> StringMassPair;

    TRed(std::vector<std::string> momenta,
         std::vector<StringMassPair> masses,
         int relprec = 8,
         int method = PV_Reduction | E_DD_Reduction);
    ~TRed() {}

    // Updates the numerical mass in the basis and updates kinematics info
    void set_mass(const std::string & name, cmplx val);
    // Sets the PSP and updates all kinematics info
    void set_momenta(const std::vector<StringMomPair> & vec);

    // Adds a needed coefficient to TRed's list, automatically adding
    // dependencies
    CoefficientNode::Ptr push_coefficient(int points,
                                          std::vector<std::string> momenta,
                                          std::vector<std::string> masses,
                                          std::vector<int> indices);

    // Point-specific coefficient requests
    CoefficientNode::Ptr add_a_coeff(std::string m, int rank, ...);
    CoefficientNode::Ptr add_b_coeff(std::string p1, std::string m1,
                                     std::string m2,
                                     int rank, ...);
    CoefficientNode::Ptr add_c_coeff(std::string p1, std::string p2,
                                     std::string m1, std::string m2,
                                     std::string m3,
                                     int rank, ...);
    CoefficientNode::Ptr add_d_coeff(std::string p1, std::string p2,
                                     std::string p3,
                                     std::string m1, std::string m2,
                                     std::string m3, std::string m4,
                                     int rank, ...);
    CoefficientNode::Ptr add_e_coeff(std::string p1, std::string p2,
                                     std::string p3, std::string p4,
                                     std::string m1, std::string m2,
                                     std::string m3, std::string m4,
                                     std::string m5, int rank, ...);
    CoefficientNode::Ptr add_f_coeff(std::string p1, std::string p2,
                                     std::string p3, std::string p4,
                                     std::string p5,
                                     std::string m1, std::string m2,
                                     std::string m3, std::string m4,
                                     std::string m5, std::string m6,
                                     int rank, ...);
    // Evaluate all needed coefficients at base precision
    bool evaluate();
    // Update mu^2 needed by scalar integrals and TRed's internal A/B evals
    void set_musq(double musq);
    // Sets the relative cutoff for threshhold tests inside BNode.
    // Used primarily for debugging.
    void set_zerocutoff(const double cutoff) { BNode::zerocutoff_ = cutoff; }

  private:
    KinematicsMap kmap_;
    ScalarIntInterface_t sif_;
    TensorIntInterface_t tif_;
    CoefficientPool pool_;
    std::vector<Poly3T<double> > results_;
    std::vector<EvaluationNode::Ptr> pushed_coeffs_;
    std::vector<StringMomPair> km_momenta_;
    std::vector<StringMassPair> km_masses_;
    bool values_set_;
};

#endif // TRED_H_
