#ifndef UTILS_H_
#define UTILS_H_

#include <algorithm>
#include <vector>
#include <limits>

/**********************************************************************
 *                                                                    *
 * Utils                                                              *
 *                                                                    *
 **********************************************************************/

namespace Utils {

// handle the passed vector as cyclic and return the element at pos.
// pos can be positive or negative
template<typename T> typename std::vector<T>::const_iterator cyclic_element_at(
    const std::vector<T> & vec, int pos) {
  if(pos >= 0 && pos < vec.size() )
    return vec.begin() + pos;
  if(pos < 0)
    return vec.begin() + (vec.size() + pos);
  return vec.begin() + (pos % vec.size());
}

// TODO: abs should be handled correctly now, eliminate redundant abs_value
// custom abs function, inconsistent abs() and fabs() definitions in STL
template<typename T> T abs_value(T x) {
  return x < 0 ? -x : x;
}

template <typename T> struct abs_less : public std::binary_function< T, T, bool>
{
  bool operator()(T x, T y) {
    return abs_value(x) < abs_value(y);
  }
};

// check for equality within a specified precision.
// precision should always be given relative to the maximal value that
// is expected!!!
template<typename T> bool is_equal(T v1, T v2, T prec) {
  return (abs_value(v1 - v2) <= prec);
}

// returns the absolute maximum value in vec
template<typename T> T max_abs_element(const std::vector<T> & vec) {
  if(vec.size() == 0)
    return 0;
#ifdef DEBUG
  std::cout << "Looking for max val in vec of size " << vec.size() << std::endl;
  for(typename std::vector<T>::const_iterator it = vec.begin();
      it != vec.end(); ++it)
    std::cout << *it << std::endl;
  std::cout << "max: "
            <<  abs_value(*std::max_element(vec.begin(),
                                            vec.end(),
                                            abs_less<T>()))
            << std::endl;
#endif
  return abs_value(*std::max_element(vec.begin(), vec.end(), abs_less<T>()));
}

// returns number of elements in vec that are within value/magnitude +- precision
template<typename T> int n_elements_equal_to(const std::vector<T> & vec,
                                             T value, T precision) {
  int count = 0;
  for(typename std::vector<T>::const_iterator it = vec.begin();
      it != vec.end(); ++it)
  if(is_equal(*it, value, precision))
    count++;
  return count;
}

// Returns the order of elements in vec such that the longest sublist of zeros
// appears on the left. The order might be reversed if this operation
// brings the largest value in the list to the right.
template<typename T> std::vector<int> order_by_zero_groups(
    const std::vector<T> & vec, T precision) {
  int group_count = -1;
  int start_index = -1;
  int largest_group_index = -1;
  int largest_group_size = -1;
  int direction = 0;
  T largest_neighbor = -std::numeric_limits<T>::max();
  typename std::vector<T>::const_iterator elem_left = vec.end();
  typename std::vector<T>::const_iterator elem_right = vec.end();

  for(std::vector<int>::size_type i = 0; i < vec.size(); ++i) {
    if(is_equal<T>(vec[i], 0., precision)) {
      group_count++;
      if(start_index == -1)
        start_index = i;
    }
    if(!is_equal<T>(vec[i], 0., precision) || i == vec.size()-1) {
      if(start_index == -1)
        continue;
      if(group_count + 1 >= largest_group_size) {
        elem_left = cyclic_element_at(vec, start_index - 1);
        elem_right = cyclic_element_at(vec, start_index + 1 + group_count);
        if((*elem_left < largest_neighbor
           && *elem_right < largest_neighbor)
           && group_count + 1 == largest_group_size)
          continue;
        largest_group_index = start_index;
        largest_group_size = group_count + 1;
        if(*elem_right <= *elem_left) {
          direction = +1;
          largest_neighbor = *elem_left;
        } else {
          direction = -1;
          largest_group_index += group_count;
          largest_neighbor = *elem_right;
        }
        group_count = -1;
        start_index = -1;
      }
    }
  }

  std::vector<int> result_indices;
  if((largest_group_index == 0 && direction == 1) || largest_group_index == -1)
    return result_indices;

  if(direction == -1) {
    for(typename std::vector<T>::size_type i = vec.size(); i != 0; --i)
      result_indices.push_back((i + largest_group_index) % vec.size());
  } else {
    for(typename std::vector<T>::size_type i = 0;  i != vec.size(); ++i)
      result_indices.push_back((i + largest_group_index) % vec.size());
  }

  return result_indices;
}

// reorders vec according to the indices given in order
template<typename T> void reorder(std::vector<T> & vec,
                                  const std::vector<int> & order) {
  if(order.size() != vec.size()) {
    std::cerr << "Error: Utils::reorder dimensions of vector and order "
              << "do not match" << std::endl;
    exit(1);
  }
  typename std::vector<T> newvec(vec.size());
  for(typename std::vector<T>::size_type i = 0; i < vec.size(); ++i)
    newvec[i] = vec[order[i]];
  vec = newvec;
}

} // namespace Utils
#endif  // UTILS_H_
